
#include "ipp4.h"
#include "pddl-parser.h"
#include "simulator.h"


/* global variables */
	
/* technical variables */
char *act_filename;
int lineno = 1;
inst_effect_list help_inst_effect_list;/* used for del_ALL_quantifiing */
BOOLEAN minimize = TRUE;/* has to do with recursive op choice in fn search */

/* globally important variables */

char ops_file[MAX_LENGTH]  = "";/* resulting name for ops file */
char fct_file[MAX_LENGTH]  = "";/* same for fct file */
char plan_file[MAX_LENGTH] = "";
char noisy_domain_file[MAX_LENGTH] = "";


/* domain constants */
op_list loaded_ops;/* loaded, uninstantiated operators */
operator_list operators;/* list of instantiated operators */
token_list initial_facts;/* the initial facts */
effect_list raw_goal_facts;/* not yet preprocessed goal facts */
token_list goal_facts; /* the goal facts */
fact_list orig_constant_list;/* to store all typed objects */
fact_list orig_initial_facts;/* stores initials as fact_list */
fact_list negated_preds = NULL;/* needed for NOT preprocessing */
fact_list types;/* fast access to all objects of one type */
token_list all_initial_facts;


/* variables for program timing */

struct tms inizio1,inizio2,fine1,fine2, beginning;

float parsing_time, ops_convertion_time;

// for external planner

int count=0;
float planner_time=0;
float planner1_time=0;


////////////////////////////////////////////////////////////////////////
//DA IPP4
//////////////////////////////////////////////////////////////////////

/* used to time the different stages of the planner */
struct tms gstart, gend, lstart, lend;
float gtotal_time = 0, gexcl_time = 0;

/* the command line inputs */
struct _command_line gcmd_line;

/* simple help: store names of connectives */
char *gconnectives[] = {"ATOM", "NOT", "AND", "OR", "ALL", "EX", "WHEN",
			 "TRU", "FAL", "EMPTY", "DUMMY", "NEUTRAL"};

/* word size of the used machine */
const int gcword_size = sizeof(int) * 8;

/* record memory consumption */
int gmemory = 0, rifo_memory = 0, ggraph_memory = 0, gexcl_memory = 0;
int gmemo_memory = 0, gwave_memory = 0;


/************ PARSING ************/

/* used for pddl parsing, flex only allows global variables */
int gbracket_count = 0;
char *gproblem_name;

/* The current input filename */
char *gact_filename;

/* The pddl domain name */
char *gdomain_name = NULL;

// loaded, uninstantiated operators

PlOperator *gloaded_ops = NULL;

/* stores initials as fact_list */
PlNode *gorig_initial_facts = NULL;

/* not yet preprocessed goal facts */
PlNode *gorig_goal_facts = NULL;

/* axioms as in UCPOP before being changed to ops */
PlOperator *gloaded_axioms = NULL;

/* the types, as defined in the domain file */
TypedList *gparse_types = NULL;

/* the constants, as defined in domain file */
TypedList *gparse_constants = NULL;

/* the predicates and their arg types, as defined in the domain file */
TypedListList *gparse_predicates = NULL;

/* the objects, declared in the problem file */
TypedList *gparse_objects = NULL;

/* connection to instantiation ( except ops, goal, initial ) */

/* all typed objects */
FactList *gorig_constant_list = NULL;

/* the predicates and their types */
FactList *gpredicates_and_types = NULL;

/* type hierarchy (PDDL)  */
type_tree_list gglobal_type_tree_list;

/* helper for types parsing */
FactList *gtypes;

/* switching between AIPS-2000 Competion Style
   and original IPP format
 */
int gofficial_output_style = FALSE;
int gnum_plan_ops;


/****************** INSTANTIATING * *****************/

/* global arrays of constant names,
 *               type names (with their constants),
 *               predicate names,
 *               predicate aritys,
 *               defined types of predicate args
 */
String gconstants_table[MAX_CONSTANTS_TABLE];
int gconstants_table_size = 0;
StringIntegers gtypes_table[MAX_TYPES_TABLE];
int gtype_size[MAX_TYPES_TABLE];
int gtypes_table_size = 0;
String gpredicates_table[MAX_PREDICATES_TABLE];
int garity[MAX_PREDICATES_TABLE];
int gpredicates_args_type[MAX_PREDICATES_TABLE][MAX_ARITY];
int gpredicates_table_size = 0;

/* the parsed input structures, translated into CodeNodes */
CodeNode *gcode_initial_state = NULL;
CodeNode *gcode_goal_state = NULL;
CodeOperator *gcode_operators = NULL;

/* helper in solving the Atomic Instantiation problem:
 *                                    the implicit tuple tables
 *
 * ( see Technical Report 122, "Handling of Inertia in a Planning System" )
 */
int_pointer gtuples[MAX_PREDICATES_TABLE];
int gone_table_size[MAX_PREDICATES_TABLE];

/* stores inertia - information: is any occurence of the predicate
 * added / deleted in the uninstantiated ops ?
 *
 * ( see TR 122 )
 */
Bool gis_added[MAX_PREDICATES_TABLE];
Bool gis_deleted[MAX_PREDICATES_TABLE];

/* store the final "relevant facts", see TR 122 */
RelevantFact_pointer grelevant_facts[MAX_RELEVANT_FACTS];
int gnum_relevant_facts = 0;
CodeOperator *ginst_code_operators = NULL;/* helper: first get all inst.ops */

/* standard name for inferred types ( unary inertia, see TR 122 ) */
char gnew_types_name[MAX_LENGTH] = "INFERRED TYPE";

/* standard name for GOAL-REACHED fact, as needed for disjunctive goals */
char ggoal_reached_name[MAX_LENGTH] = "GOAL-REACHED";

//non strettamente necessari

OpArray *gops_at;
int *gnum_ops_at;
OpArray gplan_ops;
int gft_vector_length;
int gprint_ftnum = 0, gprint_exnum = 0;
int gfirst_full_time = 0;

unsigned int gfacts_count = 0, gexclusions_count = 0;
unsigned int gops_count = 0, gops_exclusions_count = 0;

/************************** BITMAP REPRESENTATION **************************/

/* the bitvector length for relevant facts */
int gft_vector_length;

/* final representation of ops,
 *                         initial state,
 *                         goal state
 */
BitOperator *gbit_operators = NULL, *bit_op;
int gnum_bit_operators = 0;
FactInfoPair *gbit_initial_state = NULL;
FactInfoPair *gbit_goal_state = NULL;



////////////////////////////////////////////////////////////////////////
//FINE DA IPP4
////////////////////////////////////////////////////////////////////////


/* setting default values */

void initialize_preset_values(void)
{

  build_ad_time=0;              /* reset timing vars */
  fixpoint_time=0;

  max_nodes = MAX_MAX_NODES;	/* max number of nodes at one level */
  display_info = 1;		/* level of run time info printed */
  debug = 0;      		/* gpg options; description see command line or gpg.h */

};


/* setting initial values for simulation */

void initialize_simul_values(void)
{

  simul.planner.type = FF_GPG;

  simul.plan_length = MAXINT;
  simul.executed_plan_length = 0;
  simul.remaining_plan_length = MAXINT;

  simul.num_inc = 0;
  simul.num_actions = 0;

  simul.noise_rate = 10;
  simul.goal_change_rate = 20;
  simul.conn_change_rate = 50;
  simul.noise_change = INIGOAL;

  simul.max_noisy_ops = MAXINT;
  simul.applyed_noisy_ops = 0;
  simul.irrelevant_noisy_ops = 0;
  simul.not_applyable_noisy_ops = 0;
  simul.num_noise_initial_facts = 0;
  simul.num_noise_goals = 0;
  strcpy(simul.message,"");

  simul.goal_heuristics = 1;

  simul.total_execution_time=0;

}


void print_simul_options(void)
{

  printf("\nUsage of SIMULATOR\n");

  printf("\nOPTIONS   DESCRIPTIONS\n\n");
  printf("-o <str>    operator file name\n");
  printf("-f <str>    fact file name\n");
  printf("-p <str>    path for operator and fact file\n\n");

  printf("-P <str>    ### Input Plan\n");
  printf("-n <str>    ### Noisy Domain File Name\n");
  printf("-N <num>    ### Noise Rate (default=%d%)\n", simul.noise_rate);
  printf("-M <num>    ### Number of max noisy operators (default=%d)\n\n", simul.max_noisy_ops);

  printf("-I <num>    ### Change <num> Initial Facts\n");
  printf("-F <num>    ### Change <num> Goals\n\n");

  printf("-G <num>    ### Goal Heuristics \n");
  printf("     0      : Simple \n");
  printf("     1      : Causal Links (default) \n");
  printf("     2      : Causal Links + Goal Agenda \n\n");

  printf("-X <num>    ### External Planner\n");
  printf("     0      : ADJ\n");
  printf("     1      : FF\n");
  printf("     2      : IPP4\n");
  printf("     10     : FF & ADJ (default)\n\n");


  printf("NOTE: default value for operator file is 'domain.pddl',\n");
  printf("      and similarly 'noisy_domain.pddl' for noisy domain file.\n");
  printf("      All suffixes are added automatically. \n\n");

}


float print_info_parsing()
{
   float total_parsing_time;

   total_parsing_time = parsing_time+ops_convertion_time;

   printf("\n GPG: spent  %.2f  seconds loading and parsing input files", parsing_time);
   printf("\n GPG: spent  %.2f  seconds converting operators (code ops -> token list)", ops_convertion_time);
   printf("\n GPG: spent  %.2f  seconds total parsing time\n", total_parsing_time);

   return (total_parsing_time);
}


void get_path (char * full_file_name, char *result)
{

    int last_slash=-1,i;
    for (i=0; i<strlen(full_file_name); i++)
       if (full_file_name[i]=='/')
	  last_slash = i;

    if (last_slash==-1)
       strcpy(result,"./");

    else
    {
       strcpy (result, full_file_name);
       result [last_slash+1]='\0';
    }
}


void build_file_name(char* full_name, char* file_name,
                     char* file_ext, char* path)
{
   char *sExt;   // This is the extension of a file name

   sExt = file_name;
   sExt += (int) (strlen(file_name) - strlen(file_ext));
     // Get last chars of file name (the extension)
   if ( strcmp ( sExt, file_ext ) )
      strcat( file_name, file_ext );

   strcat( full_name, path );
   strcat( full_name, file_name );
}



// -------------------- Main ----------------------------

int main( int argc, char *argv[] )
{

  int exit_status = 0; /* indicates at end if plan was found, used for PURIFY */

  char path[MAX_LENGTH]  = "";/* common path for both files */
  char ops_file_name[MAX_LENGTH] = "domain.pddl";/* default name */
  char fct_file_name[MAX_LENGTH] = "";/* <fct_file_name>.pddl */
  char plan_file_name[MAX_LENGTH] = "";/* <plan_file_name>.sol */
  char noisy_domain_file_name[MAX_LENGTH] = "noisy_domain.pddl";

  int min_time = 0;/* specified minimal planning time */
  int max_time = 0;
  int curr_min_time;

  int level;
  BOOLEAN ok_plan = FALSE;/* stores return value of simulation process */
  BOOLEAN found_plan = FALSE;

  char option;/* command line interpretation */

  // for noise generation ...
  struct tms  end_time ;
  struct timeval tv;
  struct timezone tz;

  CodeOperator *o;

  char simul_path[MAX_LENGTH];

//ricava il path nel quale si trova il simulatore (puo' essere relativo o assoluto)
  get_path ( *argv, simul_path);

  /* global initialization */
  initialize_preset_values();
  initialize_simul_values();
  initialize_gpg_fixpoint();


  /* no options entered: print usage */
  if (argc == 1)
  {
     print_simul_options();
     exit( 1 );
  }


  /* read and interpret command line arguments */
  while ( --argc && ++argv )
  {

     if ( *argv[0] == '-' && (strlen(*argv) == 2) )
     {                          /* option specified */

        option = *++argv[0];
	switch ( option )
	{

        default:

        if ( --argc && ++argv )
	{
	switch ( option )
	{                        /* examine the option letter */
	  case 'o':
	    strncpy(ops_file_name, *argv, MAX_LENGTH);
	    break;
	  case 'f':
	    strncpy(fct_file_name, *argv, MAX_LENGTH);
	    break;
          case 'P':
            strncpy(plan_file_name, *argv, MAX_LENGTH);
            break;
          case 'n':
            strncpy(noisy_domain_file_name, *argv, MAX_LENGTH);
            break;
          case 'p':
	    strncpy(path, *argv, MAX_LENGTH);
            break;

	  case 'X':
            sscanf(*argv,"%d", &simul.planner.type);
	    break;
          case 'N':
            sscanf(*argv,"%d", &simul.noise_rate);
            break;
          case 'M':
            sscanf(*argv,"%d", &simul.max_noisy_ops);
            break;
          case 'G':
	    sscanf( *argv,"%d",&simul.goal_heuristics);
	    break;
          case 'I':
            sscanf( *argv,"%d",&simul.num_noise_initial_facts);
	    break;
	  case 'F':
	    sscanf( *argv,"%d",&simul.num_noise_goals);
	    break;

	  default:

	    printf( "\nSIMUL: unknown option: %c entered\n\n", option );
	    exit( 1 );
	    break;

        } // of switch
	} // if

        else
        {           /* something wrong with the option */
           printf("\nSIMUL: unknown option\n\n");
	   exit( 1 );
        }
      } // switch
    } //if

    else
    {               /* option has more than one letter */
       printf("\nSIMUL: wrong usage\n\n");
       exit( 1 );
    }
  } // while


  initialize_simul_params();

//ricava la linea di comando del planner esterno
//il path del planner esterno deve essere definito in maniera relativa a partire dal path di gpg-adj
  strcpy (SIMUL_COMM_LINE, simul_path);
  strcat (SIMUL_COMM_LINE, simul.planner.command_line);
  strcpy (SIMUL_COMM_LINE1, simul_path);
  strcat (SIMUL_COMM_LINE1, simul.planner.command_line1);


/* add type suffixes: extension for input files and path info */
  build_file_name(ops_file, ops_file_name, ".pddl", path);
  build_file_name(fct_file, fct_file_name,".pddl", path);
  build_file_name(plan_file, plan_file_name,".sol", path);
  build_file_name(noisy_domain_file, noisy_domain_file_name, ".pddl", path);


//directory temporanea per I/O con planner
  strcpy (TEMP_DIR, simul_path);
  strcat (TEMP_DIR, "temp/");

  if (mkdir (TEMP_DIR, 0777)==-1)
    if (errno != EEXIST)
    {
       printf ("\nCan't create directory %s, err %d\n\n", TEMP_DIR,errno);
       exit(1);
    }

  times (&inizio1); // loading and parsing ...

  load_ops_file( noisy_domain_file );
  load_fct_file( fct_file );
  min_time=load_plan( plan_file );

  simul.executed_plan_length = 0;
  simul.remaining_plan_length = simul.plan_length = min_time;


  printf("\nPARSING & ISTANTIATING (please wait) ... \n");

  /* This is needed to get all types */
  new_build_orig_constant_list();

  /* preprocess PlNodes so as to be able to translate them into CodeNodes */
  /* assign unique names to variables and normalize effects   */
  pddl_outer();

  /* collect all strings in the domain representation, store them into global string arrays, and
   * collect inertia information on the way. Also create types and predicate args types */

  /* compute values of global arrays
   *
   * gconstants_table
   * gtypes_table
   * gpredicates_table
   * garity
   * gpredicates_args_type
   */
  collect_all_strings();

  /* translate Input formula into coded format;
   *
   * also do syntax checking
   * (legal effects, conditions, quantification etc.)
   * and structure normalization
   */
  transform_PlNodes_to_CodeNodes();

  /* perform advanced domain definition check:
   * simplify input, i.e., search for formula that can be
   * trivially simplified and issue warnings, if one is found.
   *
   * then, find logical tautologies, like A and not A, and issue
   * warnings for any such occurence.
   *
   * afterwards, continue planning on the again simplified
   * (due to tautologies) domain description
   *
   * NOTE: the arguments to that functions only say that "yes, issue
   *       a warning wenever you find something"
   */
  simplify_all_CodeNodes( TRUE );
  detect_all_tautologies( TRUE );
  simplify_all_CodeNodes( TRUE );

  /* prepare for instantiation process:
   * build implicit tuple tables for fast access to
   * initial state information ( see Technical Report 122 )
   *
   * also used in unary to types encoding, that's why it
   * needs to be done here already
   */
  build_predicate_tables();

  /* encode unary inertia used in op and goal state description
   * as types (see TR 122), to make instantiation
   * process a good deal faster
   *
   * afterwards, find new (maybe resulting from that) tautologies,
   * simplify description and go on.
   */
  encode_unary_inertia_in_types();
  detect_all_tautologies( FALSE );
  simplify_all_CodeNodes( FALSE );

  if ( gcmd_line.display_info == 5 )
  {
    printf("\nfinal pre-instantiation initial state is:\n");
    print_CodeNode( gcode_initial_state, 0 );
    printf("\nfinal pre-instantiation goal state is:\n");
    print_CodeNode( gcode_goal_state, 0 );
    printf("\nfinal pre-instantiation operators are:");
    for ( o = gcode_operators; o; o = o->next )
      print_CodeOperator( o );
  }

  /* the actual instantiation part, first step that actually may need
   * significant computation time: create all instances of operator
   * schemata, and transform all quantifiers (ALL/EX) into con/dis junction
   *
   * representation is simultaneously cleand up where easily possible (in
   * particular, initial + goal state are fully simplified and checked),
   * as that might well save a lot of time.
   *
   * for better overview, the final clean up of operators, tautologies and (e.g.)
   * empty effects removal, is done separately in spite of this, after also setting
   * relevants.
   *
   * anyone crazy enough to try can maybe gain some efficiency by
   * messing the two (three) parts of code together. GOOD LUCK !
   *
   * something more striking: I don't really know if at all, or how many,
   *                          tautologies can result from instantiating.
   *                          maybe one can spare the tautologie search
   *                          in all instances and then, also the simplification
   *                          of condition CodeNodes in final clean up
   */
  multiply_params_and_quantifiers();

  if ( gcmd_line.display_info == 6 )
  {
    printf("\n1. step instantiated initial state is:\n");
    print_CodeNode( gcode_initial_state, 0 );
    printf("\n1. step instantiated goal state is:\n");
    print_CodeNode( gcode_goal_state, 0 );
    printf("\n1. step instantiated operators are:");
    for ( o = ginst_code_operators; o; o = o->next )
      print_CodeOperator( o );
  }

  /* now we don't need the global tuple tables anymore; so free their memory */
  free_predicate_tables();

  /* detect all facts that can, in principle, change their truth value during
   * planning. these are the facts that are relevant to the problem.
   *
   * set all other atoms to TRUE/FALSE and perform a final clean up of the
   * actions. (initial + goal are also simplified a last time)
   */

  all_initial_facts = token_list_from_CodeNode (gcode_initial_state, 1);

  collect_relevant_facts();

  detect_inst_CodeOperator_tautologies();
  simplify_all_inst_CodeOperators();

  if ( gcmd_line.display_info == 8 )
  {
    printf("\nfinal initial state is:\n");
    print_CodeNode( gcode_initial_state, 0 );
    printf("\nfinal goal state is:\n");
    print_CodeNode( gcode_goal_state, 0 );
    printf("\nfinal operators are:");
    for ( o = ginst_code_operators; o; o = o->next )
      print_CodeOperator( o );
  }

  /* transform the cleaned up and simplified CodeNode ops, initial and goal
   * into bitmaps, i.e., represent each fact by its index in grelevant_facts
   * and a set of facts by the membership vector. if goal is disjunctive,
   * add a new GOAL_REACHED fact as well as one obvious op for each disjunct.
   * On the way, melt identical effects, which means that effects that have the same
   * set of preconditions are put together. The identification of an op's
   * unconditional effects is a special case of this
   */
  generate_ini_goal_bitmap_representation();

  types = types_list_from_types_table ( gtypes_table );

  initial_facts = token_list_from_FactInfo ( gbit_initial_state->positive);
  inertial_facts = get_inertials ( all_initial_facts, initial_facts);
  goal_facts = token_list_from_FactInfo ( gbit_goal_state->positive );

  times(&fine1);


// ricopio dai code ops alle token list
  times (&inizio2);
  operators = operator_list_from_CodeOperator (ginst_code_operators);
  noisy_operators = noisy_operator_from_CodeOperator (ginst_code_operators);
  times (&fine2);


/* begin timing */
   times( &start_time );

/* display info on instantiated operators, if wanted */

   if ( display_info > 4 )
   {
      printf( "\n Original instantiated operators are:\n\n" );
      print_operators( operators );
      printf( "\n Noisy instantiated operators are:\n\n" );
      print_operators( noisy_operators );
      printf("\n" );
   }

   times(&fine1);

   parsing_time = ( float ) ( ( fine1.tms_utime  - inizio1.tms_utime  +
                                fine1.tms_stime  - inizio1.tms_stime  +
                      		fine1.tms_cutime - inizio1.tms_cutime +
			        fine1.tms_cstime - inizio1.tms_cstime	) / 100.0 ) ;

   ops_convertion_time = ( float ) ( ( fine2.tms_utime  - inizio2.tms_utime  +
                                       fine2.tms_stime  - inizio2.tms_stime  +
                                       fine2.tms_cutime - inizio2.tms_cutime +
     			               fine2.tms_cstime - inizio2.tms_cstime	) / 100.0 ) ;

   print_info_parsing();


   printf("\nSTARTING SIMULATION ...\n");


   initialise_go_op_nodes();
   clean_parsing_memory();

   actual_initial_facts = copy_token_list( initial_facts );
   actual_goal_facts = copy_token_list( goal_facts );


   // generazione della variante
   if (simul.num_noise_initial_facts || simul.num_noise_goals)
   {
      change_plan(simul.num_noise_initial_facts, simul.num_noise_goals, types, inertial_facts, actual_initial_facts, actual_goal_facts, fct_file_name);
      exit(0);
   }


// init random generator for noise generation ...
   gettimeofday(&tv, &tz);
   simul.random_seed=((tv.tv_sec & 0177 )*1000000)+ tv.tv_usec;
   srandom(simul.random_seed);

   ok_plan=FALSE;


   if ( min_time )
   {

//    compute_causal_links(store_goal_facts);

      while (!ok_plan)
      {
         found_plan = FALSE;
         curr_min_time = forward_simulation();

         if( curr_min_time == NO_INCONS )
         {
             ok_plan = TRUE;
             printf("\n Simulation succesfully terminated ...\n");
             exit_status = 0;
         }

         else
         {
            max_time=0;
            save_current_state(&actual_initial_facts);
	    found_plan = find_new_plan( &max_time, types, inertial_facts, actual_initial_facts, actual_goal_facts, fct_file);
         }

	 if (found_plan)
	 {
	     times( &end_time );

	     for (level=0; level<simul.plan_length; level++)
                free_simul_vect(level);

             if (simul.planner.type == FF_GPG)
             {
                printf("\n Current simulation time: GPG = %.2f seconds", planner1_time);
                printf("\n                          FF  = %.2f seconds\n", planner_time);
             }
             else
                printf("\n Current simulation time = %.2f seconds\n", planner_time);

             simul.plan_length = insert_plan_from_file(0);

             ok_plan = FALSE;
       	  }

     }

   }

  if (ok_plan)
     exit_status = 0;
  else
     exit_status = 1;

  times( &end_time );


  if (simul.planner.type == FF_GPG)
  {
     printf("\n Total simulation time: GPG = %.2f seconds", planner1_time);
     printf("\n                        FF  = %.2f seconds\n", planner_time);
  }
  else
     printf("\n Total simulation time = %.2f seconds\n", planner_time);


  // this is only for testing
  #ifdef TEST_SIMUL
     printf("\nAABBCC::%.2f::%.2f::%d::%d::%d::%d::%d::%d\n", planner1_time, planner_time, simul.num_actions, simul.noise_rate,
            simul.max_noisy_ops, simul.applyed_noisy_ops, simul.irrelevant_noisy_ops, simul.not_applyable_noisy_ops);
  #endif /* TEST_SIMUL */

  store_simulation(simul.executed_plan_length, fct_file_name, planner_time, planner1_time);

  printf("\nEXITING FROM SIMULATOR ... \n\n\n");
  exit( exit_status );

}

