
#include "simulator.h"
#include "ipp4.h"
#include "memory.h"
#include "pddl-parser.h"


//-------------------------------------------------------------------------------------------------
//token_list_from_CodeNode
//Crea una fact list da un PlNode
//PARAMETRI
//PlNode nodelist
//RETURN
//la token list
//-------------------------------------------------------------------------------------------------

token_list token_list_from_CodeNode (CodeNode *node, int firstcall)
{
  CodeNode *i_son;
  int i, n;
  
  char atom_string[MAX_LENGTH];
  char fact_string[MAX_LENGTH]="";

  if (firstcall)
  {
     free_token_list (base);
     tl_now = NULL;
     tl_before = NULL;
  }

  if ( !node )
  {
    // empty list
    return (token_list)NULL;
  }

  switch ( node->connective )
  {

  case AND:
    token_list_from_CodeNode( node->sons, 0 );
    if ( node->sons )
    {
      for (i_son = node->sons->next; i_son!=NULL; i_son = i_son->next) {
	token_list_from_CodeNode( i_son, 0 );
      }
    }
    break;

  case ATOM:
    sprintf(atom_string,"%s", node->predicate == -1 ? "EQ" : gpredicates_table[node->predicate]);
    strcat(fact_string, atom_string);
    n = node->predicate == -1 ? 2 : garity[node->predicate];
    for ( i=0; i<n; i++ ) {
      if ( node->arguments[i] < 0 ) {
	printf("x%d ", ((-1)*node->arguments[i])-1);
      } else {
	sprintf(atom_string,"_%s", gconstants_table[node->arguments[i]]);
	strcat(fact_string, atom_string);
      }
    }

    tl_now = new_token_list();
    tl_now->item = new_token( strlen( fact_string ) + 1 );
    strcpy(tl_now->item,fact_string);
    // L'elemento precedente (se ne esiste uno)punta all'elemento appena creato
    if (tl_before)	
       tl_before->next=tl_now;
    else
       base=tl_now;
    // aggiorna il puntatore ad elemento precedente
    tl_before=tl_now;

    // azzera la stringa per il prossimo fatto
    strcpy (fact_string,"");

    if ( node->sons ) spec_error("ATOM");
    break;

  case NOT:
      printf ("\n\nNegative predicates not yet supported.\n\n");
      exit(1);
      break;

  default:
    printf("\n***** ERROR ****");
    printf("\ntoken_list_from_CodeNode: %d > Wrong Node specifier\n", node->connective);
    exit(1);
  }

  return (base);

}


//-------------------------------------------------------------------------------------------------
//preconds_list_from_CodeNode
//Crea una fact list da un PlNode
//-------------------------------------------------------------------------------------------------

void preconds_list_from_CodeNode (CodeNode *node, operator_list op)
{
  CodeNode *i_son;
  int i, n;
  char atom_string[MAX_LENGTH];
  char fact_string[MAX_LENGTH]="";


  if ( node )
  {

  switch ( node->connective )
  {

  case AND:
    preconds_list_from_CodeNode( node->sons, op );
    if ( node->sons )
    {
      for (i_son = node->sons->next; i_son!=NULL; i_son = i_son->next)
	preconds_list_from_CodeNode( i_son, op );
    }
    break;

  case ATOM:
    sprintf(atom_string,"%s", node->predicate == -1 ? "EQ" : gpredicates_table[node->predicate]);
    strcat(fact_string, atom_string);
    n = node->predicate == -1 ? 2 : garity[node->predicate];
    for ( i=0; i<n; i++ )
    {
      if ( node->arguments[i] < 0 )
	printf("x%d ", ((-1)*node->arguments[i])-1);

      else
      {
	sprintf(atom_string,"_%s", gconstants_table[node->arguments[i]]);
	strcat(fact_string, atom_string);
      }
    }

    tl_now = new_token_list();
    tl_now->item = new_token( strlen( fact_string ) + 1 );
    strcpy(tl_now->item,fact_string);
    tl_now->next = op->preconditions;
//    printf("\n precond: %s",tl_now->item);	
    op->preconditions = tl_now;

//azzera la stringa per il prossimo fatto
    strcpy (fact_string,"");

    if ( node->sons ) spec_error("ATOM");
    break;

  case NOT:
      printf ("\n\n Negative predicates not yet supported.\n\n");
      exit(1);
      break;

  default:
    printf("\n***** ERROR ****");
    printf("\n preconds_list_from_CodeNode: %d > Wrong Node specifier\n", node->connective);
    exit(1);
  }
  }

}


//-------------------------------------------------------------------------------------------------
//op_effects_list_from_CodeNode
//Crea una effects list a partire dal CodeNode
//-------------------------------------------------------------------------------------------------

void effects_list_from_CodeNode (CodeNode *node, operator_list op, int del_eff)
{
  CodeNode *i_son;
  int i, n;
  char atom_string[MAX_LENGTH];
  char fact_string[MAX_LENGTH]="";

  token_list tl_now;

  if ( node )
  {

  switch ( node->connective )
  {

  case AND:
    effects_list_from_CodeNode( node->sons, op , 0);
    if ( node->sons )
    {
      for (i_son = node->sons->next; i_son!=NULL; i_son = i_son->next)
	effects_list_from_CodeNode( i_son, op, 0 );
    }
    break;

  case NOT:
    effects_list_from_CodeNode( node->sons, op, 1 );
    if ( node->sons->next ) spec_error("NOT");
    break;

  case ATOM:
    sprintf(atom_string,"%s", node->predicate == -1 ? "EQ" : gpredicates_table[node->predicate]);
    strcat(fact_string, atom_string);
    n = node->predicate == -1 ? 2 : garity[node->predicate];
    for ( i=0; i<n; i++ )
    {
      if ( node->arguments[i] < 0 )
	printf("x%d ", ((-1)*node->arguments[i])-1);

      else
      {
	sprintf(atom_string,"_%s", gconstants_table[node->arguments[i]]);
	strcat(fact_string, atom_string);
      }
    }

    tl_now = new_token_list();
    tl_now->item = new_token( strlen( fact_string ) + 1 );
    strcpy(tl_now->item,fact_string);
    // L'elemento precedente (se ne esiste uno)punta all'elemento appena creato
    if (del_eff==0)
    {
	tl_now->next = op->effects->add_effects;
//        printf("\n del eff: %s",tl_now->item);	
	op->effects->add_effects = tl_now;
    }
    else
    {
        tl_now->next = op->effects->del_effects;
//        printf("\n add eff: %s",tl_now->item);
        op->effects->del_effects = tl_now;
    }

    // azzera la stringa per il prossimo fatto
    strcpy (fact_string,"");

    if ( node->sons ) spec_error("ATOM");
    break;

  case WHEN:
    effects_list_from_CodeNode( node->sons, op, 0 );
    break;

  case TRU:
    effects_list_from_CodeNode( node->next, op, 0 );
    break;

  case FAL:

    break;

  case EMPTY:

    break;

  default:
    printf("\n***** ERROR ****");
    printf("\n effects_list_from_CodeNode: %d > Wrong Node specifier\n", node->connective);
    exit(1);
  }
  }
}


operator_list operator_list_from_CodeOperator (CodeOperator* base_code_op)
{
    CodeOperator* code_op;
    operator_list temp=NULL;
    operator_list base=NULL;

    char aux_string[MAX_LENGTH];

    for (code_op=base_code_op; code_op!=NULL; code_op=code_op->next)
    {
	inst_operator_string_underscores (aux_string, code_op);
	
	if ( strstr (aux_string, "NOISE-") == NULL )
	{
  	   temp=new_ground_operator( aux_string );
	   temp->effects = new_effect_list();
           preconds_list_from_CodeNode( code_op->preconds, temp );
	   effects_list_from_CodeNode( code_op->conditionals, temp, 0);
	   temp->objects=NULL;

	   temp->next=base;
	   base=temp;
	}

    }

    return base;
}


operator_list noisy_operator_from_CodeOperator (CodeOperator* base_code_op)
{
    CodeOperator* code_op;
    operator_list temp=NULL;
    operator_list base=NULL;

    char aux_string[MAX_LENGTH];

    for (code_op=base_code_op; code_op!=NULL; code_op=code_op->next)
    {
	inst_operator_string_underscores (aux_string, code_op);
	
	if ( strstr (aux_string, "NOISE-") != NULL )
	{
	   temp=new_ground_operator( aux_string );
    	   temp->effects = new_effect_list();
           preconds_list_from_CodeNode( code_op->preconds, temp );
	   effects_list_from_CodeNode( code_op->conditionals, temp, 0);
	   temp->objects=NULL;

	   temp->next=base;
	   base=temp;
	}

    }

    return base;
}


token_list token_list_from_FactInfo ( FactInfo *f )
{
    Integers *i;
    RelevantFact *ff;
    char fact_string[MAX_LENGTH]="";
    token_list tl_before = NULL;
    token_list tl_now = NULL;

    base = NULL;

    if ( !f )
    {
	return (token_list)NULL;
    }

    for ( i = f->indices; i; i = i->next )
    {
	ff = grelevant_facts[i->index];
//	printf("\n%d: ", i->index);
	fact_string_underscores( fact_string,ff->predicate, ff->arguments );

	tl_now = new_token_list();
	tl_now->item = new_token( strlen( fact_string ) + 1 );
	strcpy(tl_now->item,fact_string);
        // L'elemento precedente (se ne esiste uno)punta all'elemento appena creato
	if (tl_before)
	    tl_before->next=tl_now;
	else
	    base=tl_now;
        // aggiorna il puntatore ad elemento precedente
	tl_before=tl_now;
        // azzera la stringa per il prossimo fatto
	strcpy (fact_string,"");
    }

//    free_FactInfo(??);
    return base;
}


FactList* types_list_from_types_table (StringIntegers* gtypes_table)
{
    FactList *result=NULL, *temp;
    int i=0;
    Integers* obj;
    TokenList *new_item, *old_item;

    while ( gtypes_table[i].name != NULL)
    {
	temp = ipp4_new_fact_list();
	temp->item = ipp4_new_token_list();
	temp->item->item = ipp4_new_token( strlen( gtypes_table[i].name ) + 1 );
    	strcpy (temp->item->item, gtypes_table[i].name);
	old_item = temp->item;

	for (obj = gtypes_table[i].integers; obj; obj=obj->next)
	{
	    new_item = ipp4_new_token_list();
	    new_item->item = ipp4_new_token ( strlen( gconstants_table[obj->index] ) + 1 );
	    strcpy (new_item->item, gconstants_table[obj->index]);
	    old_item->next = new_item;
	    old_item = new_item;
	    new_item->next = NULL;
	}

	temp->next=result;
	result = temp;
	i++;
    }
    return result;
}


void fact_string_underscores (char* out_string, short int predicate, ArgArray arguments )
{
  int j;
  char aux_string[MAX_LENGTH];

  strcpy(out_string,"");
  sprintf(aux_string,"%s", gpredicates_table[predicate]);
  strcat(out_string,aux_string);

  for ( j=0; j<garity[predicate]; j++ )
  {
    sprintf(aux_string,"_%s", gconstants_table[arguments[j]]);
    strcat(out_string,aux_string);
  }
}


//Ricava il nome completo di un operatore instanziato partendo da un CodeOperator
void inst_operator_string_underscores (char* out_string, CodeOperator* op)
{
  int j;
  char aux_string[MAX_LENGTH];

  strcpy(out_string,"");
  sprintf(aux_string,"%s", op->name);
  strcat(out_string,aux_string);

  for ( j=0; j<op->num_vars; j++ )
  {
      sprintf(aux_string,"_%s",gconstants_table[op->inst_table[j]]);
      strcat(out_string,aux_string);
  }

//  printf("\n\n%s\n",out_string);

}


token_list get_inertials (token_list all_facts, token_list relevant_facts)
{

    token_list result=NULL,temp;

    BOOLEAN present;
    token_list i_tok, j_tok;

    // Scorre la lista dei fatti iniziali alla ricerca di predicati inerziali
    for (i_tok=all_facts; i_tok; i_tok=i_tok->next)
    {
	present = FALSE;
	for ( j_tok = relevant_facts; j_tok; j_tok = j_tok->next)
	{
	    if (strstr (i_tok->item, j_tok->item))
		present = TRUE;
	}
	if (!present)
	{
            // Trovato un fatto che prevede un predicato inerziale, lo aggiunge alla lista di output
	    temp = ( token_list ) calloc( 1, sizeof( token_list_elt ));
	    temp->item = i_tok->item;
	    temp->next = result;
	    result = temp;
	}
    }
    return result;
}



void clean_parsing_memory()
{
   register CodeOperator *op;

   for (op=ginst_code_operators; op; op=op->next)
      free_CodeOperator(op);

   ginst_code_operators = NULL;
/*
   if (operators != NULL)
      free_operator_list(operators);
*/
}