



/*********************************************************************
 * File: pddl-types.c
 * Description: Functions for the pddl dependend part of
 *              the parser and the preprocessing.
 *
 * Author: Frank Rittinger 1998, Joerg Hoffmann 2000
 *
 *********************************************************************/ 
/*********************************************************************
 * (C) Copyright 1998 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 *********************************************************************/



#include "ipp4.h"
#include "pddl.h"
#include "utilipp4.h"
#include "memory.h"

#include "pddl-types.h"




TokenList *
type_already_known(char * name, FactList * types)
{
  FactList * f;

  for ( f = types; f; f = f->next )
    if ( strcmp( f->item->item, name ) == SAME )
      return f->item->next;
  return NULL;
}

type_tree 
main_type_tree()
{
  type_tree_list ttl;

  for ( ttl = gglobal_type_tree_list; ttl; ttl = ttl->next )
    if ( strcmp( ttl->item->name, STANDARD_TYPE ) == SAME )
      return ttl->item;
  return NULL;
}

type_tree 
new_type_tree(char * name)
{
  type_tree act_type;
  
  if (!name)
    return NULL;
  act_type = ( type_tree ) calloc( 1, sizeof( type_tree_elt ) );
  CHECK_PTR(act_type);
  act_type->name = ipp4_new_token( strlen( name ) + 1 );
  strcpy( act_type->name, name );
  act_type->sub_types = NULL;
  return act_type;
}

type_tree_list 
new_type_tree_list(char * name)
{
  type_tree_list act_type_list;
  
  act_type_list = ( type_tree_list ) calloc( 1, sizeof( type_tree_list_elt ) );
  CHECK_PTR(act_type_list);
  if ( name )
    act_type_list->item = new_type_tree( name );
  else
    act_type_list->item = NULL;
  act_type_list->next = NULL;
  
  return act_type_list;
}

/* steps recursively through type tree and searches for name */
type_tree 
find_branch(char * name, type_tree root )
{
  type_tree p;
  type_tree_list ttl;

  if ( !root )
    return NULL;
  if ( strcmp( root->name, name ) == SAME )
    return root;
  if ( !root->sub_types )
    return NULL;
  for ( ttl=root->sub_types; ttl; ttl=ttl->next )
    if ((p = find_branch( name, ttl->item )))
      return p;
  return NULL; 
}

/* calls itself recursively to get all objects that are of the types and
   subtypes of ttl */
FactList *
build_object_list_from_ttl(type_tree_list ttl, FactList * types_done )
{
  FactList * f;
  FactList * td;
  FactList * std;
  TokenList * t;
  TokenList * tl_dummy;
  TokenList * st;
  type_tree_list sttl;

  if ( !ttl )
    return types_done;
  
  types_done = build_object_list_from_ttl( ttl->next, types_done );

  if ((t = type_already_known( ttl->item->name, types_done)))
    return types_done;

  td = ipp4_new_fact_list();
  t = td->item = ipp4_new_token_list();
  /* begin with the name of the type... */
  t->item = copy_string( ttl->item->name );
  for ( f = gorig_constant_list; f; f = f->next )
    {
      /* ...followed by objects of that type. */
      if ( strcmp( f->item->next->item, ttl->item->name ) == SAME )
	{
	  t->next = ipp4_new_token_list();
	  t = t->next;
	  t->item = copy_string( f->item->item );
	}
    }
  /* now append the objects of the subtypes */
  std =  build_object_list_from_ttl( ttl->item->sub_types, types_done );
  /* now we can be sure that for each subtype a list with all
     objects of that type is somewhere in std. We simply take these
     lists and copy all of them into a new one for the supertype */
  for ( sttl = ttl->item->sub_types; sttl; sttl = sttl->next )
    {
      st = type_already_known( sttl->item->name, std );
      if ( st )
	{
	  t->next = copy_complete_token_list( st, &tl_dummy );
	  t = tl_dummy;
	}
    }
  td->next = std;
  return td;
}

void 
build_orig_constant_list()
{
  FactList * f;
  FactList * nextf;
  FactList * end = NULL;
  TokenList * t;
  Bool do_count;
  int objects_count = 0;

  gtypes = build_object_list_from_ttl( gglobal_type_tree_list, 
						 NULL );
  free_complete_fact_list( gorig_constant_list );
  gorig_constant_list = NULL;
  for ( f = gtypes; f; f = nextf )
    {
      nextf = f->next;
      if ( strcmp( f->item->item, STANDARD_TYPE ) == SAME )
	do_count = TRUE;
      else
	do_count = FALSE;
      for ( t = f->item->next; t; t = t->next )
	{
	  if ( !gorig_constant_list )
	    gorig_constant_list = end = ipp4_new_fact_list();
	  else
	    {
	      end->next = ipp4_new_fact_list();
	      end = end->next;
	    }
	  end->item = ipp4_new_token_list();
	  end->item->item = copy_string( t->item );
	  end->item->next = ipp4_new_token_list();
	  end->item->next->item = copy_string( f->item->item );
	  if ( do_count )
	    {
	      /* count objects for RIFO meta strategy */
	      objects_count++;
	    }
	}
    }
}

void
add_to_type_tree(FactList * t_list, type_tree tree )
{
  type_tree branch = tree;
  type_tree_list new_branch;
  char * this_type; 
  char * super_type;

  /* step through list and build a hierarchy of types */
  for( ; t_list; t_list=t_list->next )
    {
      this_type = t_list->item->item;
      if ( !t_list->item->next ) 
	{
	  fprintf(stderr, "\n%s: error at '%s'.\n", gact_filename, this_type );
	  exit( PARSE_ERROR_CODE );
	}
      super_type = t_list->item->next->item;
      if ( strcmp( branch->name, super_type ) != SAME )
	branch = find_branch( super_type, tree );
      if ( !branch ) 
	{
	  fprintf(stderr, "\n%s: unknown type '%s'.\n", 
		  gact_filename, super_type );
	  exit( PARSE_ERROR_CODE );
	}
      /* now the type is a subtype of the one currently looked at 
	 in the type tree */
      new_branch = new_type_tree_list( this_type );
      new_branch->next = branch->sub_types;
      branch->sub_types = new_branch;
    }
}

void 
print_type_tree_list( type_tree_list root, int indent )
{
  int i;
  for(i=0;i<2*indent; i++ ) 
    {
      fprintf(stdout, " " );
    }
  if ( root ) 
    {
      if ( !root->item || !root->item->name )
	{
	  fprintf(stderr, "\n%s: internal error: no type name specified.\n", 
		  gact_filename );
	  exit( INTERNAL_ERROR_CODE );
	}
      else
	fprintf(stdout, "%s\n", root->item->name );
      if ( root->item->sub_types )
	print_type_tree_list( root->item->sub_types, indent+1 );
      if ( root->next )
	print_type_tree_list( root->next, indent );
    }
}

/********************************************************************* 
 * Insert this subtree into the preconditions. Take care whether there
 * are 0, 1 or >1 preconditions and 0, 1, >1 eq-predicates. Prepend a 
 * AND node if necessary.
 *
 * PlOperator * op: The current operator. 
 * PlNode * root: The root of the new subtree that is inserted.
 *
 *********************************************************************/
void 
insert_not_eq(PlOperator * op, PlNode * root)
{
    PlNode * tmp;

    if (NULL == op->preconds) {
	op->preconds = root;
    } else {
	if (AND == op->preconds->connective) {
	    /* There is already a AND node, simply append the tree. */
	    tmp = op->preconds->sons;
	    op->preconds->sons = root;
	    root->next = tmp;
	} else {
	    /* There is no AND. First create a new AND, prepend it to
	       the preconds and append the subtree. */
	    tmp = new_pl_node(AND);
	    tmp->sons = root;
	    root->next = op->preconds;
	    op->preconds = tmp;
	}
    }
}

/*********************************************************************
 * This function goes over all operators and inserts a "not="
 * predicate for every two parameters in one operator that have the
 * same types.
 *
 * PlOperator * ops: The global operator list.
 *
 *********************************************************************/
void 
adjust_equal_types(PlOperator * ops)
{
    PlOperator * tmp_op;
    FactList * tmp_out;
    FactList * tmp_in;
    PlNode * tmp_node;

    for (tmp_op = ops; tmp_op; tmp_op = tmp_op->next) {
	for (tmp_out = tmp_op->params; tmp_out; tmp_out = tmp_out->next) {
	    for (tmp_in = tmp_out->next; tmp_in; tmp_in = tmp_in->next) {
		/* If the types of two parameters are the same, 
		   create a new "not=" predicate and insert it
		   appropriately (AND or no AND). */
		if (tmp_out->item->next->item ==
		    tmp_in->item->next->item || SAME ==
		    strcmp(tmp_out->item->next->item,
			   tmp_in->item->next->item)) {
#ifdef MYDEBUG
		    fprintf(stdout, "\n>>> OP: %s \nSAME: %s == %s",
			    tmp_op->name->item, tmp_out->item->item,
			    tmp_in->item->item); 
#endif		    
		    /* Create the new subtree for the "not=". */
		    tmp_node = new_pl_node(NOT);
		    tmp_node->sons = new_pl_node(ATOM);
		    tmp_node->sons->atom = ipp4_new_token_list();
		    tmp_node->sons->atom->item = EQ_STR;
		    tmp_node->sons->atom->next = ipp4_new_token_list();
		    tmp_node->sons->atom->next->item =
			ipp4_new_token(strlen(tmp_out->item->item) + 1);
		    strcpy(tmp_node->sons->atom->next->item,
			   tmp_out->item->item);
		    tmp_node->sons->atom->next->next = ipp4_new_token_list();
		    tmp_node->sons->atom->next->next->item =
			ipp4_new_token(strlen(tmp_in->item->item) + 1);
		    strcpy(tmp_node->sons->atom->next->next->item,
			   tmp_in->item->item);
#ifdef MYDEBUG
		    fprintf(stdout, " --> Creating new negated Atom: ");
		    print_tokenlist(tmp_node->sons->atom, " ");
#endif
		    /* Insert this subtree into the preconditions.
		       Take care whether there are 0, 1 or >1
		       preconditions and 0, 1, >1
		       eq-predicates. Prepend a AND node if
		       necessary. */
		    insert_not_eq(tmp_op, tmp_node);
		}
	    }
	}
    }
}













/*********************************************
 * NEW IMPLEMENTATION OF TYPES PREPROCESSING *
 * (wenn man net alles selber macht, gelle ? *
 *********************************************/











/* typed-list-of   preprocessing
 */







Token ltype_names[MAX_TYPES_TABLE];
int lnum_types;


int leither_ty[MAX_TYPES_TABLE][MAX_TYPES_TABLE];
int lnum_either_ty[MAX_TYPES_TABLE];





void new_build_orig_constant_list( void )

{

  char *tmp = NULL;
  TypedList *tyl;
  TypedListList *tyll;
  TokenList *tl, *p_tl, *tmp_tl;
  PlOperator *po;

  int i, j, k, n, std;

  Bool m[MAX_TYPES_TABLE][MAX_TYPES_TABLE];

  FactList *fl, *p_fl;


  lnum_types = 0;
  for ( tyl = gparse_types; tyl; tyl = tyl->next ) {
    if ( my_get_type( tyl->name ) == -1 ) {
      ltype_names[lnum_types++] = copy_Token( tyl->name );
    }
    if ( tyl->type->next ) {
      tmp = ipp4_new_token( MAX_LENGTH );
      strcpy( tmp, EITHER_STR );
      for ( tl = tyl->type; tl; tl = tl->next ) {
	strcat( tmp, CONNECTOR );
	strcat( tmp, tl->item );
      }
    } else {
      tmp = copy_Token( tyl->type->item );
    }
    if ( (n = my_get_type( tmp )) == -1 ) {
      tyl->n = lnum_types;
      ltype_names[lnum_types++] = copy_Token( tmp );
    } else {
      tyl->n = n;
    }
    free( tmp );
    tmp = NULL;
  }
     
  for ( tyl = gparse_constants; tyl; tyl = tyl->next ) {
    if ( tyl->type->next ) {
      tmp = ipp4_new_token( MAX_LENGTH );
      strcpy( tmp, EITHER_STR );
      for ( tl = tyl->type; tl; tl = tl->next ) {
	strcat( tmp, CONNECTOR );
	strcat( tmp, tl->item );
      }
    } else {
      tmp = copy_Token( tyl->type->item );
    }
    if ( (n = my_get_type( tmp )) == -1 ) {
      tyl->n = lnum_types;
      ltype_names[lnum_types++] = copy_Token( tmp );
    } else {
      tyl->n = n;
    }
    free( tmp );
    tmp = NULL;
  }
  
  for ( tyl = gparse_objects; tyl; tyl = tyl->next ) {
    if ( tyl->type->next ) {
      tmp = ipp4_new_token( MAX_LENGTH );
      strcpy( tmp, EITHER_STR );
      for ( tl = tyl->type; tl; tl = tl->next ) {
	strcat( tmp, CONNECTOR );
	strcat( tmp, tl->item );
      }
    } else {
      tmp = copy_Token( tyl->type->item );
    }
    if ( (n = my_get_type( tmp )) == -1 ) {
      tyl->n = lnum_types;
      ltype_names[lnum_types++] = copy_Token( tmp );
    } else {
      tyl->n = n;
    }
    free( tmp );
    tmp = NULL;
  }

  for ( tyll = gparse_predicates; tyll; tyll = tyll->next ) {
    for ( tyl = tyll->args; tyl; tyl = tyl->next ) {
      if ( tyl->type->next ) {
	tmp = ipp4_new_token( MAX_LENGTH );
	strcpy( tmp, EITHER_STR );
	for ( tl = tyl->type; tl; tl = tl->next ) {
	  strcat( tmp, CONNECTOR );
	  strcat( tmp, tl->item );
	}
      } else {
	tmp = copy_Token( tyl->type->item );
      }
      if ( (n = my_get_type( tmp )) == -1 ) {
	tyl->n = lnum_types;
	ltype_names[lnum_types++] = copy_Token( tmp );
      } else {
	tyl->n = n;
      }
      free( tmp );
      tmp = NULL;
    }
  }
    
  collect_type_names_in_pl( gorig_goal_facts );

  for ( po = gloaded_ops; po; po = po->next ) {
    collect_type_names_in_pl( po->preconds );
    collect_type_names_in_pl( po->effects );
    for ( tyl = po->parse_params; tyl; tyl = tyl->next ) {
      if ( tyl->type->next ) {
	tmp = ipp4_new_token( MAX_LENGTH );
	strcpy( tmp, EITHER_STR );
	for ( tl = tyl->type; tl; tl = tl->next ) {
	  strcat( tmp, CONNECTOR );
	  strcat( tmp, tl->item );
	}
      } else {
	tmp = copy_Token( tyl->type->item );
      }
      if ( (n = my_get_type( tmp )) == -1 ) {
	tyl->n = lnum_types;
	ltype_names[lnum_types++] = copy_Token( tmp );
      } else {
	tyl->n = n;
      }
      free( tmp );
      tmp = NULL;
    }
  }


  /* now get the numbers of all composed either types
   */
  for ( i = 0; i < lnum_types; i++ ) {
    lnum_either_ty[i] = 0;
  }
  for ( tyl = gparse_types; tyl; tyl = tyl->next ) {
    make_either_ty( tyl );
  }
  for ( tyl = gparse_constants; tyl; tyl = tyl->next ) {
    make_either_ty( tyl );
  }
  for ( tyl = gparse_objects; tyl; tyl = tyl->next ) {
    make_either_ty( tyl );
  }
  for ( tyll = gparse_predicates; tyll; tyll = tyll->next ) {
    for ( tyl = tyll->args; tyl; tyl = tyl->next ) {
      make_either_ty( tyl );
    }
  }
  make_either_ty_in_pl( gorig_goal_facts );
  for ( po = gloaded_ops; po; po = po->next ) {
    make_either_ty_in_pl( po->preconds );
    make_either_ty_in_pl( po->effects );
    for ( tyl = po->parse_params; tyl; tyl = tyl->next ) {
      make_either_ty( tyl );
    }
  }


  /* now, compute the transitive closure of all type inclusions.
   * first initialize the matrix.
   */
  for ( i = 0; i < lnum_types; i++ ) {
    for ( j = 0; j < lnum_types; j++ ) {
      m[i][j] = ( i == j ? TRUE : FALSE );
    }
  }
  std = -1;
  for ( i = 0; i < lnum_types; i++ ) {
    if ( strcmp( ltype_names[i], STANDARD_TYPE ) == SAME ) {
      std = i;
      break;
    }
  }
  for ( i = 0; i < lnum_types; i++ ) {
    m[i][std] = TRUE;/* all types are subtypes of OBJECT */
  }
  for ( tyl = gparse_types; tyl; tyl = tyl->next ) {
    /* all inclusions as are defined in domain file
     */
    m[my_get_type( tyl->name )][tyl->n] = TRUE;
  }
  /* compute transitive closure on inclusions matrix
   */
  for ( j = 0; j < lnum_types; j++ ) {
    for ( i = 0; i < lnum_types; i++ ) {
      if ( m[i][j] ) {
	for ( k = 0; k < lnum_types; k++ ) {
	  if ( m[j][k] ) {
	    m[i][k] = TRUE;
	  }
	}
      }
    }
  }
  /* union types are subsets of all those types that contain all
   * their components, and 
   * all component types are subsets of the either type !
   */
  for ( i = 0; i < lnum_types; i++ ) {
    if ( lnum_either_ty[i] < 2 ) continue;
    for ( j = 0; j < lnum_types; j++ ) {
      if ( j == i ) continue;
      /* get supertypes of all component types
       */
      for ( k = 0; k < lnum_either_ty[i]; k++ ) {
	if ( !m[leither_ty[i][k]][j] ) break;
      }
      if ( k < lnum_either_ty[i] ) continue;
      m[i][j] = TRUE;
      /* make components subtypes of either type
       */
      for ( k = 0; k < lnum_either_ty[i]; k++ ) {
	m[leither_ty[i][k]][i] = TRUE;
      }
    }
  }
  /* and again, compute transitive closure on inclusions matrix.
   * I guess, this won't change anything (?), but it also won't need
   * any remarkable computation time, so why should one think about it ?
   */
  for ( j = 0; j < lnum_types; j++ ) {
    for ( i = 0; i < lnum_types; i++ ) {
      if ( m[i][j] ) {
	for ( k = 0; k < lnum_types; k++ ) {
	  if ( m[j][k] ) {
	    m[i][k] = TRUE;
	  }
	}
      }
    }
  }
  

  /* now build FactList of ALL  constant -> type   pairs.
   * for each constant / object, let it appear separately
   * for each type it is a member of; compute type
   * membership based on propagating constants / objects
   * through inclusions matrix.
   *
   * this might make the same pair appear doubly, if an object
   * is declared in type T as well as in some supertype T'.
   * such cases will be filtered out in string collection.
   */
  for ( tyl = gparse_constants; tyl; tyl = tyl->next ) {
    fl = ipp4_new_fact_list();
    fl->item = ipp4_new_token_list();
    fl->item->next = ipp4_new_token_list();
    fl->item->item = copy_Token( tyl->name );
    if ( tyl->type->next ) {
      fl->item->next->item = ipp4_new_token( MAX_LENGTH );
      strcpy( fl->item->next->item, EITHER_STR );
      for ( tl = tyl->type; tl; tl = tl->next ) {
	strcat( fl->item->next->item, CONNECTOR );
	strcat( fl->item->next->item, tl->item );
      }
    } else {
      fl->item->next->item = copy_Token( tyl->type->item );
    }
    fl->next = gorig_constant_list;
    gorig_constant_list = fl;
    /* now add constant to all supertypes
     */
    n = my_get_type( fl->item->next->item );
    for ( i = 0; i < lnum_types; i++ ) {
      if ( i == n ||
	   !m[n][i] ) continue;
      fl = ipp4_new_fact_list();
      fl->item = ipp4_new_token_list();
      fl->item->next = ipp4_new_token_list();
      fl->item->item = copy_Token( tyl->name );
      fl->item->next->item = copy_Token( ltype_names[i] );
      fl->next = gorig_constant_list;
      gorig_constant_list = fl;
    }
  }
  for ( tyl = gparse_objects; tyl; tyl = tyl->next ) {
    fl = ipp4_new_fact_list();
    fl->item = ipp4_new_token_list();
    fl->item->next = ipp4_new_token_list();
    fl->item->item = copy_Token( tyl->name );
    if ( tyl->type->next ) {
      fl->item->next->item = ipp4_new_token( MAX_LENGTH );
      strcpy( fl->item->next->item, EITHER_STR );
      for ( tl = tyl->type; tl; tl = tl->next ) {
	strcat( fl->item->next->item, CONNECTOR );
	strcat( fl->item->next->item, tl->item );
      }
    } else {
      fl->item->next->item = copy_Token( tyl->type->item );
    }
    fl->next = gorig_constant_list;
    gorig_constant_list = fl;
    /* now add constant to all supertypes
     */
    n = my_get_type( fl->item->next->item );
    for ( i = 0; i < lnum_types; i++ ) {
      if ( i == n ||
	   !m[n][i] ) continue;
      fl = ipp4_new_fact_list();
      fl->item = ipp4_new_token_list();
      fl->item->next = ipp4_new_token_list();
      fl->item->item = copy_Token( tyl->name );
      fl->item->next->item = copy_Token( ltype_names[i] );
      fl->next = gorig_constant_list;
      gorig_constant_list = fl;
    }
  }


  /* now, normalize all typed-list-of  s in domain and problem def,
   * i.e., in all PlNode quantifiers and in op parameters
   *
   * at the same time, remove typed-listof structures in these defs
   */
  normalize_tyl_in_pl( &gorig_goal_facts );
  for ( po = gloaded_ops; po; po = po->next ) {
    normalize_tyl_in_pl( &po->preconds );
    normalize_tyl_in_pl( &po->effects );
    /* be careful to maintain parameter ordering !
     */
    if ( !po->parse_params ) {
      continue;/* no params at all */
    }
    fl = ipp4_new_fact_list();
    fl->item = ipp4_new_token_list();
    fl->item->next = ipp4_new_token_list();
    fl->item->item = copy_Token( po->parse_params->name );
    if ( po->parse_params->type->next ) {
      fl->item->next->item = ipp4_new_token( MAX_LENGTH );
      strcpy( fl->item->next->item, EITHER_STR );
      for ( tl = po->parse_params->type; tl; tl = tl->next ) {
	strcat( fl->item->next->item, CONNECTOR );
	strcat( fl->item->next->item, tl->item );
      }
    } else {
      fl->item->next->item = copy_Token( po->parse_params->type->item );
    }
    po->params = fl;
    p_fl = fl;
    for ( tyl = po->parse_params->next; tyl; tyl = tyl->next ) {
      fl = ipp4_new_fact_list();
      fl->item = ipp4_new_token_list();
      fl->item->next = ipp4_new_token_list();
      fl->item->item = copy_Token( tyl->name );
      if ( tyl->type->next ) {
	fl->item->next->item = ipp4_new_token( MAX_LENGTH );
	strcpy( fl->item->next->item, EITHER_STR );
	for ( tl = tyl->type; tl; tl = tl->next ) {
	  strcat( fl->item->next->item, CONNECTOR );
	  strcat( fl->item->next->item, tl->item );
	}
      } else {
	fl->item->next->item = copy_Token( tyl->type->item );
      }
      p_fl->next = fl;
      p_fl = fl;
    }
    free_TypedList( po->parse_params );
    po->parse_params = NULL;
  }


  /* finally, build  gpredicates_and_types  by chaining predicate names 
   * together with the names of their args' types.
   */
  for ( tyll = gparse_predicates; tyll; tyll = tyll->next ) {
    fl = ipp4_new_fact_list();
    fl->item = ipp4_new_token_list();
    fl->item->item = copy_Token( tyll->predicate );
    fl->next = gpredicates_and_types;
    gpredicates_and_types = fl;
    if ( !tyll->args ) continue;
    /* add arg types; MAINTAIN ORDERING !
     */
    fl->item->next = ipp4_new_token_list();
    if ( tyll->args->type->next ) {
      fl->item->next->item = ipp4_new_token( MAX_LENGTH );
      strcpy( fl->item->next->item, EITHER_STR );
      for ( tl = tyll->args->type; tl; tl = tl->next ) {
	strcat( fl->item->next->item, CONNECTOR );
	strcat( fl->item->next->item, tl->item );
      }
    } else {
      fl->item->next->item = copy_Token( tyll->args->type->item );
    }
    p_tl = fl->item->next;
    for ( tyl = tyll->args->next; tyl; tyl = tyl->next ) {
      tmp_tl = ipp4_new_token_list();
      if ( tyl->type->next ) {
	tmp_tl->item = ipp4_new_token( MAX_LENGTH );
	strcpy( tmp_tl->item, EITHER_STR );
	for ( tl = tyl->type; tl; tl = tl->next ) {
	  strcat( tmp_tl->item, CONNECTOR );
	  strcat( tmp_tl->item, tl->item );
	}
      } else {
	tmp_tl->item = copy_Token( tyl->type->item );
      }
      p_tl->next = tmp_tl;
      p_tl = tmp_tl;
    }
  }


  /* now get rid of remaining typed-list-of parsing structures
   */
  free_TypedList( gparse_types );
  gparse_types = NULL;
  free_TypedList( gparse_constants );
  gparse_constants = NULL;
  free_TypedList( gparse_objects );
  gparse_objects = NULL;
  free_TypedListList( gparse_predicates );
  gparse_predicates = NULL;

}



void collect_type_names_in_pl( PlNode *n )

{

  PlNode *i;
  TypedList *tyl;
  TokenList *tl;
  char *tmp = NULL;
  int nn;

  if ( !n ) {
    return;
  }

  switch( n->connective ) {
  case ALL:
  case EX:
    for ( tyl = n->parse_vars; tyl; tyl = tyl->next ) {
      if ( tyl->type->next ) {
	tmp = ipp4_new_token( MAX_LENGTH );
	strcpy( tmp, EITHER_STR );
	for ( tl = tyl->type; tl; tl = tl->next ) {
	  strcat( tmp, CONNECTOR );
	  strcat( tmp, tl->item );
	}
      } else {
	tmp = copy_Token( tyl->type->item );
      }
      if ( (nn = my_get_type( tmp )) == -1 ) {
	tyl->n = lnum_types;
	ltype_names[lnum_types++] = copy_Token( tmp );
      } else {
	tyl->n = nn;
      }
      free( tmp );
      tmp = NULL;
    }
    collect_type_names_in_pl( n->sons );
    break;
  case AND:
  case OR:
    for ( i = n->sons; i; i = i->next ) {
      collect_type_names_in_pl( i );
    }
    break;
  case NOT:
    collect_type_names_in_pl( n->sons );
    break;
  case ATOM:
  case TRU:
  case FAL:
    break;
  case WHEN:
    collect_type_names_in_pl( n->sons );
    collect_type_names_in_pl( n->sons->next );
    break;
  default:
    break;
  }

}



int my_get_type( char *str )

{

  int i;

  for ( i = 0; i < lnum_types; i++ ) {
    if ( strcmp( str, ltype_names[i] ) == SAME ) return i;
  }

  return -1;

}



void make_either_ty( TypedList *tyl )

{

  TokenList *i;

  if ( lnum_either_ty[tyl->n] > 0 ) {
    return;
  }

  for ( i = tyl->type; i; i = i->next ) {
    leither_ty[tyl->n][lnum_either_ty[tyl->n]++] = my_get_type( i->item );
  }

}



void make_either_ty_in_pl( PlNode *n )

{

  PlNode *i;
  TypedList *tyl;

  if ( !n ) {
    return;
  }

  switch( n->connective ) {
  case ALL:
  case EX:
    for ( tyl = n->parse_vars; tyl; tyl = tyl->next ) {
      make_either_ty( tyl );
    }
    make_either_ty_in_pl( n->sons );
    break;
  case AND:
  case OR:
    for ( i = n->sons; i; i = i->next ) {
      make_either_ty_in_pl( i );
    }
    break;
  case NOT:
    make_either_ty_in_pl( n->sons );
    break;
  case ATOM:
  case TRU:
  case FAL:
    break;
  case WHEN:
    make_either_ty_in_pl( n->sons );
    make_either_ty_in_pl( n->sons->next );
    break;
  default:
    break;
  }

}



void normalize_tyl_in_pl( PlNode **n )

{

  PlNode *i;
  TypedList *tyl;
  PlNode *tmp_pl = NULL, *sons, *p_pl;
  TokenList *tmp_tl, *tl;


  if ( !(*n) ) {
    return;
  }

  switch( (*n)->connective ) {
  case ALL:
  case EX:
    /* we need to make a sequence of quantifiers ( ->sons ...)
     * out of the given sequence of TypedList  elements,
     * with connected type names, var - name in TokenList
     * and KEEPING THE SAME ORDERING !!
     */
    if ( !(*n)->parse_vars ) {
      printf("\n\nquantifier without argument !! check input files.\n\n");
      exit( 1 );
    }
    tmp_tl = ipp4_new_token_list();
    tmp_tl->next = ipp4_new_token_list();
    tmp_tl->item = copy_Token( (*n)->parse_vars->name );
    if ( (*n)->parse_vars->type->next ) {
      tmp_tl->next->item = ipp4_new_token( MAX_LENGTH );
      strcpy( tmp_tl->next->item, EITHER_STR );
      for ( tl = (*n)->parse_vars->type; tl; tl = tl->next ) {
	strcat( tmp_tl->next->item, CONNECTOR );
	strcat( tmp_tl->next->item, tl->item );
      }
    } else {
      tmp_tl->next->item = copy_Token( (*n)->parse_vars->type->item );
    }
    (*n)->atom = tmp_tl;
    /* now add list of sons
     */
    sons = (*n)->sons;
    p_pl = *n;
    for ( tyl = (*n)->parse_vars->next; tyl; tyl = tyl->next ) {
      tmp_tl = ipp4_new_token_list();
      tmp_tl->next = ipp4_new_token_list();
      tmp_tl->item = copy_Token( tyl->name );
      if ( tyl->type->next ) {
	tmp_tl->next->item = ipp4_new_token( MAX_LENGTH );
	strcpy( tmp_tl->next->item, EITHER_STR );
	for ( tl = tyl->type; tl; tl = tl->next ) {
	  strcat( tmp_tl->next->item, CONNECTOR );
	  strcat( tmp_tl->next->item, tl->item );
	}
      } else {
	tmp_tl->next->item = copy_Token( tyl->type->item );
      }
      tmp_pl = new_pl_node( (*n)->connective );
      tmp_pl->atom = tmp_tl;
      p_pl->sons = tmp_pl;
      p_pl = tmp_pl;
    }
    /* remove typed-list-of info
     */
    free_TypedList( (*n)->parse_vars );
    (*n)->parse_vars = NULL;
    /* the last son in list takes over ->sons
     */
    p_pl->sons = sons;
    /* normalize this sons and get out
     */
    normalize_tyl_in_pl( &(p_pl->sons) );
    break;
  case AND:
  case OR:
    for ( i = (*n)->sons; i; i = i->next ) {
      normalize_tyl_in_pl( &i );
    }
    break;
  case NOT:
    normalize_tyl_in_pl( &((*n)->sons) );
    break;
  case ATOM:
  case TRU:
  case FAL:
    break;
  case WHEN:
    normalize_tyl_in_pl( &((*n)->sons) );
    normalize_tyl_in_pl( &((*n)->sons->next) );
    break;
  default:
    break;
  }

}











