

#include "simulator.h"

extern int count;
extern float planner_time;
extern float planner1_time;

extern char ops_file[];
extern char fct_file[];



void initialize_simul_params()
{
  switch (simul.planner.type)
  {
     case GPG :
        strcpy(simul.planner.command_line, GPG_COMM_LINE);
        break;
     case FF :
        strcpy(simul.planner.command_line, FF_COMM_LINE);
        break;
     case IPP4 :
        strcpy(simul.planner.command_line, IPP4_COMM_LINE);
        break;
     case FF_GPG :
        strcpy(simul.planner.command_line, FF_COMM_LINE);
        strcpy(simul.planner.command_line1, GPG_COMM_LINE);
        break;

  }
}


// initialise g_table and o_table without building fixpoint

void initialise_go_op_nodes( void )
{

  token_list tl;
  operator_list o;
  inst_effect_list e;
  o_node_list t_o;
  e_node_list t_e;
  g_node_list go;

  int index;

  for ( tl = initial_facts; tl; tl = tl->next )
  {
    go = insert_go( tl );
    go->is_initial = TRUE;
  }

  for ( o = operators; o; o = o->next )
    for ( e = o->effects; e; e = e->next )
      for ( tl = e->add_effects; tl; tl = tl->next )
	go = insert_go( tl );

  for ( o = operators; o; o = o->next )
  {
    if ( applicable( o->preconditions ) )
    {
      t_o = ( o_node_list ) calloc( 1, sizeof( o_node_list_elt ) );
      t_o->name = o->name;
      t_o->preconds = NULL;
      for ( tl = o->preconditions; tl; tl = tl->next )
      {
	go = quick_lookup_go( tl );
	t_o->preconds = insert_g_edge( go, t_o->preconds );
	go->preconds = insert_o_edge( t_o, go->preconds );
      }
      t_o->effects = NULL;
      for ( e = o->effects; e; e = e->next )
      {
	if ( applicable( e->conditions ) )
	{
	  t_e = ( e_node_list ) calloc( 1, sizeof( e_node_list_elt ) );
	  t_e->on = TRUE;
	  t_e->op = t_o;
	  t_e->conditions = NULL;
	  for ( tl = e->conditions; tl; tl = tl->next )
	  {
	    go = quick_lookup_go( tl );
	    t_e->conditions = insert_g_edge( go, t_e->conditions );
	    go->effconds = insert_e_edge( t_e, go->effconds );
	  }
	  t_e->add_edges = NULL;
	  for ( tl = e->add_effects; tl; tl = tl->next )
	  {
	    go = quick_lookup_go( tl );
	    t_e->add_edges = insert_g_edge( go, t_e->add_edges );
	    go->add_edges = insert_e_edge( t_e, go->add_edges );
	  }
	  t_e->del_edges = NULL;
	  for ( tl = e->del_effects; tl; tl = tl->next )
	  {
	    go = quick_lookup_go( tl );
	    if ( go )
	    {
	      t_e->del_edges = insert_g_edge( go, t_e->del_edges );
	      go->del_edges = insert_e_edge( t_e, go->del_edges );
	    }
	  }
	  t_e->next = t_o->effects;
	  t_o->effects = t_e;
	}
      }

      // inserting operators into o_table ...
      index = hash_index( t_o->name );
      t_o->next = o_table[index];
      o_table[index] = t_o;
      set_gam_op_uid( t_o , gam_total_num_ops );
      gam_total_num_ops++;

    }
  }
}


// Initialization for gpg and fix point

void initialize_gpg_fixpoint()
{
  int i;

  for ( i = 0; i<G_HSIZE; i++ )
  {
     g_table[i] = NULL;
     o_table[i] = NULL;
  }

  gam_total_num_ops = 0;
  gam_total_num_goals = 0;

  gam_fixpoint_reached = FALSE;
  gam_fixpoint_level = 0;

  bytes_o = 0;
  bytes_f = 0;
  bytes_e = 0;
  bytes_k = 0;
  bytes_excl = 0;
  numbereff = 0;
  numberkan = 0;
  numberexclkan = 0;

  bytes_graph = 0;
  bytes_memo = 0;

}


// Clean g_table variables used in forward simulation

void init_g_table()
{
    int i;
    g_node_list go;

    for ( i = 0; i < G_HSIZE; i++ )
      for ( go = g_table[i]; go; go = go->next )
      {
	  go->gpg.is_true = FALSE;
 	  go->gpg.is_precond = FALSE;
	  go->gpg.is_added = FALSE;
	  go->gpg.is_deleted = FALSE;
          go->gpg.is_goal = FALSE;
      }
}


int hash_index( char *key )
{
  register int h;
  int index;

  for( h=0; *key != '\0'; key++ )
     h = ( h * 256 + ( *key ) ) % BIGPRIME;

  index = ( int ) ( h % G_HASH );
//   index = ( int ) ( h & G_HASH );

  return (index);
}


g_node_list lookup_go( token t )
{

  g_node_list g;

  for ( g = g_table[ hash_index( t ) ]; g; g = g->next )
    if ( strcmp( g->name, t ) == SAME ) break;

  if ( g )
    return g;
  else
    return NULL;
}


o_node_list lookup_op( token t )
{

  o_node_list o;

  for ( o = o_table[ hash_index( t ) ]; o; o = o->next )
    if ( strcmp( o->name, t ) == SAME ) break;

  if ( o )
    return o;
  else
    return NULL;

}


g_node_list quick_lookup_go( token_list tl )
{
   if ( tl->go_pointer == NULL )
      tl->go_pointer = lookup_go (tl->item);

   return (tl->go_pointer);
}


o_node_list quick_lookup_act( simul_action* act )
{
   if ( act->op_pointer == NULL )
      act->op_pointer = lookup_op (act->name);

   return (act->op_pointer);
}


g_node_list insert_go( token_list tl )
{

  g_node_list node;
  int index;

  if ( ( node = quick_lookup_go( tl ) ) != NULL )
     return (node);

  if ( gam_total_num_goals == MAX_MAX_NODES )
  {
     printf("\ngam: too many facts. increase MAX_MAX_NODES, currently %d\n\n", MAX_MAX_NODES);
     exit( 1 );
  }

  if ( debug == 7 )
     printf("\ninserting goal %s now\n", tl->item);

  index = hash_index( tl->item );

  node = ( g_node_list ) calloc( 1, sizeof( g_node_list_elt ) );
  CHECK_MEMORY( node );
  GRAPHF(g_node_list_elt);

  node->name = new_token( strlen( tl->item ) + 1 );
  strcpy(node->name, tl->item);

  node->false_facts = NULL;
  node->true_facts = NULL;

  node->Tcount = 0;
  node->is_true = FALSE;
  node->st = 0;
  node->hits = 0;
  node->uid_block = 0;
  node->uid_mask = 0;
  node->is_initial = FALSE;
  node->count = 0;
  node->preconds = NULL;
  node->effconds = NULL;
  node->add_edges = NULL;
  node->del_edges = NULL;

  node->next = g_table[index];
  g_table[index] = node;
  tl->go_pointer = node;

  set_gam_go_uid( node , gam_total_num_goals );
  gam_total_num_goals++;

  return (node);

}


// Print critic informations about a goal node

void print_goal(g_node_list go, BOOLEAN details)
{
   printf("\n - %s ",go->name);

   if (details)
   {
      printf(" (");
      printf(" is_true: %d ", go->gpg.is_true);
      printf(" is_precond: %d ", go->gpg.is_precond);
      printf(" is_added: %d ", go->gpg.is_added);
      printf(" is_deleted: %d ", go->gpg.is_deleted);
      printf(" is_goal: %d ", go->gpg.is_goal);
      printf(")\n");
   }
}


void print_g_table()
{

   int i;
   g_node_list go;

   BOOLEAN goal_details = TRUE;

   printf("\n Contenuto della g_table :\n");

   for ( i = 0; i < G_HSIZE; i++ )
     for ( go = g_table[i]; go; go = go->next )
         print_goal(go,goal_details);

   printf("\n");

}



void restore_current_state(token_list tl)
{

  token_list temp_ft;

  init_g_table();

  for( temp_ft = tl; temp_ft; temp_ft = temp_ft->next )
  {
      temp_ft->go_pointer->gpg.is_true=TRUE;
  }

}



void save_current_state(token_list* tl)
{

  int i;
  token_list temp;
  g_node_list go;

  free_token_list(*tl);
  *tl=NULL;

  for ( i = 0; i < G_HSIZE; i++ )
     for ( go = g_table[i]; go; go = go->next )
     {
        if (go->gpg.is_true)
        {

           temp = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
           temp->item = new_token( strlen( go->name ) + 1 );
           strcpy(temp->item, go->name);

           temp->go_pointer = go;
           temp->next = *tl;
           *tl = temp;

        }
     }
}



// Count not satisfied goals at level "time"

int goal_check(int time)
{

  int not_satisfied = 0;

  token_list temp_go;
  g_node_list go;

  for ( temp_go = actual_goal_facts; temp_go; temp_go = temp_go->next )
  {
    go = quick_lookup_go ( temp_go );

    if (!go->gpg.is_true)
    {
       printf("\n SIMUL: Goal '%s' not satisfied at level %d", go->name, time);
       not_satisfied++;
    }
  }

  return (not_satisfied);
}



// Update fields in g_table

int update_facts()
{

  int i;
  g_node_list go;

  token_list temp;
  free_token_list(actual_initial_facts);
  actual_initial_facts=NULL;
  /* we identificate new true facts */
  /* positive and negative effects are inserted, other g_table variables are cleaned */

  for ( i = 0; i < G_HSIZE; i++ )
    for ( go = g_table[i]; go; go = go->next )
    {

       // inconsistence found ...
       if ((go->gpg.is_precond && !go->gpg.is_true) ||
           (go->gpg.is_added && go->gpg.is_deleted))
       {
	   printf("\n SIMUL: Inconsistence found : %s ",go->name);

	   if ( debug == 3 )
	   {
 	      print_goal(go, TRUE);
              printf("\n");
           }

           return (FALSE);
       }

       if (go->gpg.is_added)
       {
           go->gpg.is_true = TRUE;
           go->gpg.is_added = FALSE;

       }
       if (go->gpg.is_deleted)
       {
           go->gpg.is_true = FALSE;
           go->gpg.is_deleted = FALSE;

       }

       if (go->gpg.is_true)
       {
           temp = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
           temp->item = new_token( strlen( go->name ) + 1 );
           strcpy(temp->item, go->name);

           temp->go_pointer = go;
           temp->next = actual_initial_facts;
           actual_initial_facts = temp;
       }

       go->gpg.is_precond = FALSE;

    }



  return (TRUE);
}


// Update goal list

int update_goals()
{

  int i;
  g_node_list go;
  token_list temp;

  /* we identificate new goal facts */
  /* positive and negative effects are inserted, other g_table variables are cleaned */
  free_token_list(actual_goal_facts);
  actual_goal_facts=NULL;


  for ( i = 0; i < G_HSIZE; i++ )
    for ( go = g_table[i]; go; go = go->next )
    {

       if (go->gpg.is_added)
       {
           go->gpg.is_goal = TRUE;
           go->gpg.is_added = FALSE;

       }
       if (go->gpg.is_deleted)
       {
           go->gpg.is_goal = FALSE;
           go->gpg.is_deleted = FALSE;

       }
       if (go->gpg.is_goal)
       {
           temp = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
           temp->item = new_token( strlen( go->name ) + 1 );
           strcpy(temp->item, go->name);

           temp->go_pointer = go;
           temp->next = actual_goal_facts;
           actual_goal_facts = temp;
       }

       go->gpg.is_precond = FALSE;

    }

  return (TRUE);
}


// returns true if every facts in list is true

BOOLEAN all_satisfied( token_list facts )
{

  g_node_list a_fact;
  token_list i_token;

  for ( i_token = facts; i_token; i_token = i_token->next )
  {
    a_fact = quick_lookup_go( i_token );

    if ( !a_fact || !a_fact->gpg.is_true )
    {
       return FALSE;
    }
  }

  return TRUE;

}


// returns true if almost a facts in list is goal

BOOLEAN fact_is_goal( token_list facts )
{

  g_node_list a_fact;
  token_list i_token;

  for ( i_token = facts; i_token; i_token = i_token->next )
  {
    a_fact = quick_lookup_go( i_token );

    if ( a_fact && a_fact->gpg.is_goal )
    {
       return TRUE;
    }
  }

  return FALSE;

}


// Apply an action in the plan

int apply_operator(simul_action* temp_act)
{

  int num_actions=0;

  o_node_list an_action;
  g_edge_list prec;
  e_node_list an_effect;
  g_edge_list add_effect, del_effect;

  an_action = quick_lookup_act(temp_act);

  if ( an_action==NULL )
  {
     if ( debug == 3 )
        printf("\n SIMUL: Action '%s' not found ! \n", temp_act->name);
  }
  else
     num_actions++;


  /* preconditions */

  for (prec = an_action->preconds; prec; prec = prec->next)
  {
      prec->endpt->gpg.is_precond = TRUE;
  }

  /* effects */

  for (an_effect = an_action->effects; an_effect ; an_effect = an_effect->next)
  {

     /* conditional effects */

        if (an_effect->conditions != NULL)
        {
           printf("\n SIMUL: Error, unable to use conditional effects !\n");
           exit(1);
        }


     /* positive effects */

        for (add_effect = an_effect->add_edges; add_effect ; add_effect = add_effect->next)
        {
           add_effect->endpt->gpg.is_added = TRUE;
        }

     /* negative effects */

        for (del_effect = an_effect->del_edges; del_effect ; del_effect = del_effect->next)
        {
 	   del_effect->endpt->gpg.is_deleted = TRUE;
        }
  }

  simul.total_execution_time++;
  return (num_actions);

}


// Choose and apply a noisy operator

BOOLEAN apply_noisy_operator()
{

  BOOLEAN applyed=FALSE;

  int i, rnd, noise_count=0;
  int noise_conn_count=0, noise_conn_tried=0, num_conn_total=0;
  operator_list temp_op;
  operator_list selected_noise, applicable_noise[MAX_NOISE], applicable_conn_noise[MAX_NOISE];
  simul_action* temp_act;

  token_list noisename;
  token_list temp_prec;
  token_list temp_add_effect;
  token_list temp_del_effect;
  inst_effect_list temp_effect;

  g_node_list prec;
  g_node_list add_effect, del_effect;

  o_node_list an_action;
  g_node_list a_fact;
  g_edge_list a_prec;


  // setting is_goal flags for all problem goals and every precondition

  for ( i = 0; i < simul.plan_length; i++)
  {

     for (temp_act = simul_vect[i]; temp_act; temp_act=temp_act->next)
     {
         an_action = quick_lookup_act(temp_act);

         for (a_prec = an_action->preconds; a_prec; a_prec = a_prec->next)
         {
            a_prec->endpt->gpg.is_goal = TRUE;
         }

     }
  }

  for ( temp_prec=actual_goal_facts; temp_prec; temp_prec=temp_prec->next )
  {

     a_fact = quick_lookup_go ( temp_prec );

     if ( a_fact == NULL )
     {
        printf("\n SIMUL: Goal fact '%s' not found! (applying noise)\n", temp_prec->item);
        continue;
     }

     a_fact->gpg.is_goal = TRUE;
  }

  // create the graph for Dijkstra Algorithm
  create_graph();

  for (temp_op=noisy_operators; temp_op; temp_op=temp_op->next)
  {

     if (strstr(temp_op->name,CONNECTION)!=NULL)
        num_conn_total++;

     if ( all_satisfied(temp_op->preconditions) && fact_is_goal(temp_op->effects->del_effects) )
     {
        noise_conn_tried++;

        if (strstr(temp_op->name,CONNECTION)!=NULL)
        {
           noisename=token_list_from_string(temp_op->name);

           if ( dijkstra(noisename->next->item, noisename->next->next->item) )
           {
              applicable_conn_noise[noise_conn_count] = temp_op;
              noise_conn_count++;
              if (noise_conn_count >= MAX_NOISE)
                break;
           }
        }
        else
        {
           applicable_noise[noise_count] = temp_op;
           noise_count++;
           if (noise_count >= MAX_NOISE)
              break;
        }
     }
  }


  // reset goal flags ...
  for ( i = 0; i < G_HSIZE; i++ )
     for ( a_fact = g_table[i]; a_fact; a_fact = a_fact->next )
     {
        a_fact->gpg.is_goal = FALSE;
     }


  // if exist almost an applicable noisy operator
  if (noise_count > 0 || noise_conn_count >0)
  {

     if (random()%100<simul.conn_change_rate)
     {
       if (noise_conn_count>0)
       {
          rnd = random();
          rnd = rnd % noise_conn_count;
          selected_noise = applicable_conn_noise[rnd];
          printf("\n Selected CONN to DELETE (%d of %d applicable) : '%s' (num_noise=%d)\n",
                  rnd, noise_conn_count, selected_noise->name, simul.applyed_noisy_ops+1);
          insert_executed_vect(selected_noise->name, FALSE);
          applyed = TRUE;
       }
       else if (noise_conn_count==0)
       {
          printf("\n CONNECTION NOISE NOT APPLYED : ANOTHER NOISE OPERATOR SELECTED");
          rnd = random();
          rnd = rnd % noise_count;
          selected_noise = applicable_noise[rnd];
          printf("\n Selected noise (%d of %d applicable) : '%s' (num_noise=%d)\n",
                  rnd, noise_conn_count, selected_noise->name, simul.applyed_noisy_ops+1);
          insert_executed_vect(selected_noise->name, FALSE);
          applyed = TRUE;
       }

     }
     else
     {
        if (noise_count>0)
        {
           rnd = random();
           rnd = rnd % noise_count;
           selected_noise = applicable_noise[rnd];
           printf("\n Selected noise (%d of %d applicable) : '%s' (num_noise=%d)\n",
                   rnd, noise_conn_count, selected_noise->name, simul.applyed_noisy_ops+1);
           insert_executed_vect(selected_noise->name, FALSE);
           applyed = TRUE;
        }
	else if (noise_count ==0)
	{
	   printf("\n NOISE NOT APPLICABLE: ANOTHER NOISE OPERATOR SELECTED");
	   rnd = random();
           rnd = rnd % noise_conn_count;
           selected_noise = applicable_conn_noise[rnd];
           printf("\n Selected CONN to DELETE (%d of %d applicable) : '%s' (num_noise=%d)\n",
                   rnd, noise_conn_count, selected_noise->name, simul.applyed_noisy_ops+1);
           insert_executed_vect(selected_noise->name, FALSE);
           applyed = TRUE;
        }
     }

     /* preconditions */

     for (temp_prec = selected_noise->preconditions; temp_prec; temp_prec = temp_prec->next)
     {
        prec = lookup_go (temp_prec->item);
        prec->gpg.is_precond = TRUE;
     }

     /* effects */

     for (temp_effect = selected_noise->effects; temp_effect ; temp_effect = temp_effect->next)
     {

        /* conditional effects */

        if (temp_effect->conditions != NULL)
        {
           printf("\n SIMUL: Error, unable to use conditional effects !\n");
           exit(1);
        }


        /* positive effects */

	for (temp_add_effect = temp_effect->add_effects; temp_add_effect ; temp_add_effect = temp_add_effect->next)
	{
	   add_effect = quick_lookup_go (temp_add_effect);
           add_effect->gpg.is_added = TRUE;
        }

        /* negative effects */

        for (temp_del_effect = temp_effect->del_effects; temp_del_effect ; temp_del_effect = temp_del_effect->next)
        {
	   del_effect = quick_lookup_go (temp_del_effect);
 	   del_effect->gpg.is_deleted = TRUE;
        }

     }
  }
  else
  {
     applyed = FALSE;
     printf("\n Sorry, no noise applicable here ... \n");
  }

  return (applyed);
}



// Look for a noisy operator that could invalidate the plan

// Choose and apply a noisy operator that change almost a goal in the problem


BOOLEAN apply_noisy_goal()
{

  BOOLEAN applyed=FALSE;

  int rnd, noise_count=0;
  operator_list temp_op;
  operator_list selected_noise, applicable_noise[MAX_NOISE];

  token_list temp_prec;
  token_list temp_add_effect;
  token_list temp_del_effect;
  inst_effect_list temp_effect;

  g_node_list prec;
  g_node_list add_effect, del_effect;

  g_node_list a_fact;



  for ( temp_prec=actual_goal_facts; temp_prec; temp_prec=temp_prec->next )
  {

     a_fact = quick_lookup_go ( temp_prec );

     if ( a_fact == NULL )
     {
        printf("\n SIMUL: Goal fact '%s' not found! (applying noise)\n", temp_prec->item);
        continue;
     }

     a_fact->gpg.is_goal = TRUE;
  }


  for (temp_op=noisy_operators; temp_op; temp_op=temp_op->next)
  {

     if ( all_satisfied(temp_op->preconditions) && fact_is_goal(temp_op->effects->del_effects) )
     {
        applicable_noise[noise_count] = temp_op;
        noise_count++;
        if (noise_count >= MAX_NOISE)
           break;
     }
  }


  // if exist almost an applicable noisy operator
  if (noise_count > 0)
  {
     rnd = random();
     rnd = rnd % noise_count;

     selected_noise = applicable_noise[rnd];
     printf("\n Selected noise (%d of %d applicable) : '%s' (num_noise=%d)\n",
                rnd, noise_count, selected_noise->name, simul.applyed_noisy_ops+1);
     insert_executed_vect(selected_noise->name, TRUE);
     applyed = TRUE;


     /* preconditions */

     for (temp_prec = selected_noise->preconditions; temp_prec; temp_prec = temp_prec->next)
     {
        prec = lookup_go (temp_prec->item);
        prec->gpg.is_precond = TRUE;
     }

     /* effects */

     for (temp_effect = selected_noise->effects; temp_effect ; temp_effect = temp_effect->next)
     {

        /* conditional effects */

        if (temp_effect->conditions != NULL)
        {
           printf("\n SIMUL: Error, unable to use conditional effects !\n");
           exit(1);
        }


        /* positive effects */

	for (temp_add_effect = temp_effect->add_effects; temp_add_effect ; temp_add_effect = temp_add_effect->next)
	{
	   add_effect = quick_lookup_go (temp_add_effect);
           add_effect->gpg.is_added = TRUE;
        }

        /* negative effects */

        for (temp_del_effect = temp_effect->del_effects; temp_del_effect ; temp_del_effect = temp_del_effect->next)
        {
	   del_effect = quick_lookup_go (temp_del_effect);
 	   del_effect->gpg.is_deleted = TRUE;
        }

     }
  }
  else
  {
     applyed = FALSE;
     printf("\n Sorry, no noise applicable here (changing goals) ... \n");
  }

  return (applyed);

}



// Forward propagation of facts : it founds inconsistences and counts the actions applyed

int forward_check()
{

  int incons_level = NO_INCONS;
  BOOLEAN ok_incons;

  int i, not_satisfied_goals;
  simul_action* temp_act;

  printf("\n Checking for inconsistences (due to noise) from current level to %d ... \n", simul.plan_length);

  save_current_state(&actual_initial_facts);


  for ( i = 0; i < simul.plan_length; i++)
  {

     for (temp_act = simul_vect[i]; temp_act; temp_act=temp_act->next)
     {
        apply_operator(temp_act);
     }

     ok_incons = update_facts();

     if ( !ok_incons )
     {
        incons_level=i;
        break;
     }
  }

  /* goals check for last level */

  if ( i==simul.plan_length )
  {
     not_satisfied_goals = goal_check(i);

     if ( not_satisfied_goals > 0 )
     {
        ok_incons = FALSE;
        incons_level = i;
     }
  }

  if ( !ok_incons )
     printf("\n\n SIMUL: Found an inconsistence at time %d (due to noise) \n\n", incons_level);
  else
     printf("\n SIMUL: No inconsistences found (due to noise) \n");

  restore_current_state(actual_initial_facts);

  return (incons_level);

}



// Forward simulation : it founds inconsistences and counts the actions applyed

int forward_simulation()
{

  int incons_level = NO_INCONS;
  BOOLEAN ok_incons, noise_applyed=FALSE;
  int rnd;
  int goal_change_rate = 100-simul.goal_change_rate;

  int i, not_satisfied_goals;
  simul_action* temp_act;


  restore_current_state(actual_initial_facts);

  printf("\n Starting forward simulation from current level to %d ... \n", simul.plan_length);


  for ( i = 0; i < simul.plan_length; i++)
  {

     for ( temp_act = simul_vect[0]; temp_act; temp_act = simul_vect[0] )
     {

        rnd=random();

        if ( ( (rnd % 100) < simul.noise_rate ) && simul.applyed_noisy_ops < simul.max_noisy_ops )
        {

           if (simul.noise_change == INITIAL || (simul.noise_change != GOAL &&
                                                ((random()% 100)  < goal_change_rate) ))
           {
              noise_applyed = apply_noisy_operator();
              update_facts();
           }
           else
           {
              noise_applyed = apply_noisy_goal();
              update_goals();
           }


           if (noise_applyed)
           {
               incons_level = forward_check();

               if ( incons_level != NO_INCONS )
               {
                  ok_incons = FALSE;
                  simul.applyed_noisy_ops++;
                  return(i);
               }
               else
                  simul.irrelevant_noisy_ops++;
           }
           else
              simul.not_applyable_noisy_ops++;
        }

        else
        {
           simul.num_actions += apply_operator(temp_act);
           insert_executed_vect(temp_act->name, FALSE);
           remove_simul_vect(temp_act);
        }

        ok_incons = update_facts();
        if ( !ok_incons ) break;

     }

     if ( !ok_incons )
     {
        incons_level = i;
        break;
     }

  }

  /* goals check for last level */

  if ( i==simul.plan_length )
  {

     not_satisfied_goals = goal_check(i);

     if ( not_satisfied_goals > 0 )
     {
        ok_incons = FALSE;
        incons_level = i;
     }

  }

  if ( !ok_incons )
     printf("\n\n SIMUL: Found an inconsistence at time %d \n\n", incons_level);

  return(incons_level);

}



void print_plan(int time_init, int time_end)
{

  int i;          /* index for time */
  BOOLEAN first;  /* for better printing: is it the first one this level? */

  simul_action* temp_act;

  /* for all non empty operator levels */
  for ( i=time_init; i<time_end; )
  {

     printf("time step %3d: ", i);
     first = TRUE;

     if (simul_vect[i]==NULL)
        printf("\t%s\n", "(no actions)");

     /* for all operators in plan */
     for(temp_act = simul_vect[i]; temp_act; temp_act = temp_act->next)
     {

	if ( first )
        {
           printf("\t%s\n", temp_act->name );
           first = FALSE;
        }
        else
	   printf("\t\t%s\n", temp_act->name );
     }

     i++;
  }

  printf( "\n" );

}


void print_facts_and_goals(int curr_init_time, int curr_max_time)
{

   printf("\nINITIAL FACTS   (time %d) \n", curr_init_time);
   print_list(initial_facts);

   printf("\nGOALS           (time %d) \n", curr_max_time);
   print_list(goal_facts);
   printf("\n");
}


// Printing of a token list : for video output

void print_list(token_list list)
{

  token_list i_token;

  for ( i_token = list; i_token; i_token = i_token->next )
    printf( "\n - %s ", i_token->item );

  printf("\n");

}



BOOLEAN applicable( token_list list )
{

  token_list t;
  g_node_list g;

  for ( t = list; t; t = t->next )
    if ( ( g = quick_lookup_go( t ) ) == NULL ) break;

  if ( t )
    return FALSE;
  else
    return TRUE;

}


g_edge_list insert_g_edge( g_node_list g, g_edge_list list )
{

  g_edge_list temp;

  temp = ( g_edge_list ) calloc( 1, sizeof( g_edge_list_elt ) );
  CHECK_MEMORY( temp );
  GRAPHK(g_edge_list_elt);
  temp->endpt = g;
  temp->next = list;

  return temp;
}


o_edge_list insert_o_edge( o_node_list o, o_edge_list list )
{

  o_edge_list temp;

  temp = ( o_edge_list ) calloc( 1, sizeof( o_edge_list_elt ) );
  CHECK_MEMORY( temp );
  GRAPHK(o_edge_list_elt);
  temp->endpt = o;
  temp->next = list;

  return temp;
}


e_edge_list insert_e_edge( e_node_list e, e_edge_list list )
{

  e_edge_list temp;

  temp = ( e_edge_list ) calloc( 1, sizeof( e_edge_list_elt ) );
  CHECK_MEMORY( temp );
  GRAPHK(e_edge_list_elt);
  temp->endpt = e;
  temp->next = list;

  return temp;
}


void print_ef_node( e_node_list e )
{

  g_edge_list g;

  printf( "con:\n" );
  for ( g = e->conditions; g; g = g->next )
    printf( "%s\n", g->endpt->name );
  printf( "add:\n" );
  for ( g = e->add_edges; g; g = g->next )
    printf( "%s\n", g->endpt->name );
  printf( "del:\n" );
  for ( g = e->del_edges; g; g = g->next )
    printf( "%s\n", g->endpt->name );

}


void print_go_op_nodes( void )
{

  o_node_list o;
  g_node_list g;
  e_node_list e;
  g_edge_list g_e;
  o_edge_list o_e;
  e_edge_list e_e;
  int i;

  for ( i = 0; i < G_HSIZE; i++ )
  {
    for ( o = o_table[i]; o; o = o->next )
    {
      printf( "operator %s:\n", o->name );
      printf( "pre:\n" );
      for ( g_e = o->preconds; g_e; g_e = g_e->next )
	printf( "%s\n", g_e->endpt->name );
      printf( "eff:\n" );
      for ( e = o->effects; e; e = e->next )
	print_ef_node( e );
    }
  }

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next )
    {
      printf( "goal: %s\npre:\n", g->name );
      for ( o_e = g->preconds; o_e; o_e = o_e->next )
	printf( "%s\n", o_e->endpt->name );
      printf( "add:\n" );
      for ( e_e = g->add_edges; e_e; e_e = e_e->next )
	print_ef_node( e_e->endpt );
      printf( "del:\n" );
      for ( e_e = g->del_edges; e_e; e_e = e_e->next )
	print_ef_node( e_e->endpt );
    }
}


void set_gam_go_uid( g_node_list g , int id )
{
  g->uid_block = id / 32;
  g->uid_mask = 1 << ( id % 32 );
}


void set_gam_op_uid( o_node_list o , int id )
{
  o->uid_block = id / 32;
  o->uid_mask = 1 << ( id % 32 );
}


//-------------------------------------------------------------------------------------------------
//search_plan_with_ext_plan
//Esegue la ricerca di un piano tramite planner esterno
//PARAMETRI
//f_list	lista dei fatti iniziali
//g_list	lista degli opiettivi
//RETURN: 1 se  stato trovato un piano, zero altrimenti
//-------------------------------------------------------------------------------------------------
int find_new_plan (int *max_time, fact_list objects_list, token_list inertia_list,
                   token_list f_list, token_list g_list, char* fct_file)
{
    char DomainFileName[MAX_LENGTH],
         ProblemFileName[MAX_LENGTH],
         ResultFileName[MAX_LENGTH],
         PlanFileName[MAX_LENGTH],
         SimulDirName[MAX_LENGTH];

    char CommandLine[10*MAX_LENGTH];
    char CommandLine1[10*MAX_LENGTH];

    char Command[10*MAX_LENGTH];
    char Command1[10*MAX_LENGTH];


    int ret_val;
    float elapsed_time = 0;

    // Indica la posizione del binario da eseguire
    strcpy(Command, SIMUL_COMM_LINE);
    strcpy(Command1, SIMUL_COMM_LINE1);

    //directory temporanea di salvataggio simulazione
    strcpy (SIMUL_DIR, TEMP_DIR);
    sprintf (SimulDirName, "simul%d/", simul.random_seed);
    strcat (SIMUL_DIR, SimulDirName);

    if (mkdir (SIMUL_DIR, 0777)==-1)
      if (errno != EEXIST)
      {
         printf ("\nCan't create directory %s, err %d\n\n", TEMP_DIR,errno);
         exit(1);
      }


    // Vengono assegnati dei nomi con numerazione progressiva
    count++;
    strcpy (DomainFileName, ops_file );
    sprintf (ProblemFileName,"%s%s%d.pddl", SIMUL_DIR,PROBLEM_FILE_PREFIX, count); //file problema (fatti)
    sprintf (ResultFileName, "%s%s%d.pddl", SIMUL_DIR,RESULT_FILE_PREFIX, count);  //file di output del planner esterno
    sprintf (PlanFileName, "%s%s%d.sol", SIMUL_DIR,PLAN_FILE_PREFIX, count);       //file del piano da adattare


    switch (simul.planner.type)
    {

      case GPG:
         sprintf (CommandLine,"%s -o %s -f %s -P %s -O %s -N 1 -G %d -c %s > /dev/null", Command, DomainFileName,
                  ProblemFileName, PlanFileName, ResultFileName, simul.goal_heuristics, fct_file);
         break;

      case FF:
         sprintf (CommandLine,"%s -o %s -f %s -O %s > /dev/null", Command, DomainFileName, ProblemFileName, ResultFileName);
      break;

      case IPP4:
         sprintf (CommandLine,"%s -o %s -f %s -O %s > /dev/null", Command, DomainFileName, ProblemFileName, ResultFileName);
      break;

      case FF_GPG:

         sprintf (CommandLine,"%s -o %s -f %s -O %s > /dev/null", Command, DomainFileName, ProblemFileName, ResultFileName);
         sprintf (CommandLine1,"%s -o %s -f %s -P %s -O %s -N 1 -G %d -c %s > /dev/null", Command1, DomainFileName,
                  ProblemFileName, PlanFileName, ResultFileName, simul.goal_heuristics, fct_file);

      break;

    }


    // Generazione dei file di input per il planner esterno
    generate_pddl_input (DomainFileName, ProblemFileName, objects_list, inertia_list, f_list, g_list);
    generate_plan_to_adapt (PlanFileName);

    // Chiamata planner esterno
    printf ("Planner Command Line: %s\n",CommandLine);
    fflush(stdout);
    call_planner (CommandLine, &elapsed_time);
    planner_time = planner_time + elapsed_time;
    // mette in una lista il piano appena trovato
    ret_val = get_plan_size (ResultFileName, max_time);
    printf ("Found? %d in %d levels\n", ret_val, *max_time);

    if (simul.planner.type == FF_GPG)
    {
       printf ("\nPlanner Command Line: %s\n",CommandLine1);
       fflush(stdout);
       call_planner (CommandLine1, &elapsed_time);
       planner1_time = planner1_time + elapsed_time;
       // mette in una lista il piano appena trovato
       ret_val = get_plan_size (ResultFileName, max_time);
       printf ("Found? %d in %d levels\n", ret_val, *max_time);
    }


    return (ret_val);
}

//-------------------------------------------------------------------------------------------------
//generate_pddl_input
//Genera i file in formato pddl per ff
//PARAMETRI:
//DomFileName	nome file dominio
//ProbFileName	nome file problema
//facts		fatti iniziali
//goals		obiettivi
//RETURN:
//sempre 1
//-------------------------------------------------------------------------------------------------
int generate_pddl_input (char* DomFileName, char* ProbFileName, fact_list objects_list, token_list inertia_list, token_list facts, token_list goals)
{
    FILE *ProblemFile;

//Apre il file in scrittura e controlla l'esito dell'operazione
    ProblemFile = fopen (ProbFileName, "w");
    if (ProblemFile == NULL)
    {
	printf ("\nCan't open file %s for writing\n", ProbFileName);
	exit (2);
    }

//Scrive il file del problema
    write_problem (ProblemFile, objects_list, inertia_list, facts, goals);

    return (1);
}


int generate_plan_to_adapt (char* PlanFileName)
{

  int i;
  FILE* fp;
  simul_action* temp_act;

  if((fp = fopen( PlanFileName, "w")) == NULL)
  {
      printf("\nCan't open file %s for writing\n", PlanFileName);
      exit(2);
  }

  /* for all non empty operator levels */
  for ( i=0; i<simul.plan_length; i++ )
  {
     fprintf(fp,"%3d: ", i);

     /* for all operators in plan */
     for (temp_act = simul_vect[i]; temp_act; temp_act = temp_act->next)
     {
        fprintf(fp,"\t%s\n", temp_act->name );
     }

  }

  fclose(fp);
  return (1);
}


//-------------------------------------------------------------------------------------------------
//write_problem
//descrizione degli oggetti del dominio
//PARAMETRI
//ProbFile	Puntatore al file del dominio
//object_list	Lista degli oggetti e loro tipo
//inertia_list	Lista fatti non modificabili
//facts		Fatti del problema corrente
//goals		Lista obiettivi del problema
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem (FILE *ProbFile, fact_list objects_list, token_list inertia_list, token_list facts, token_list goals)
{

//copia dell'header del problema originale
    write_problem_header (fct_file,ProbFile);

//Scrive intestazione fatti iniziali
    fprintf (ProbFile,"\t(:init\n");

//Scrive i fatti non modificabili da alcun operatore
    write_problem_inertias (ProbFile, inertia_list);

//Scrive sul file del problema la descrizione dell'istante iniziale ( un insieme di relazioni)
    write_problem_initial_facts (ProbFile,facts);

//Scrive sul file del problema la descrizione del goal ( anch'esso un insieme di relazioni)
    write_problem_goal_facts (ProbFile,goals);

//Definizione del problema conclusa, chiude il file

    fprintf(ProbFile,"; %s",simul.message);
    fclose (ProbFile);

    return 1;
}


//-------------------------------------------------------------------------------------------------
//call_planner
//chiamata al planner esterno ff
//PARAMETRI
//commline	linea di comando
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int call_planner (char* commline, float *elapsed_time)
{
    struct tms start_planner_time, end_planner_time;

    int ret_value;

    //Salva i tempi prima della chiamata
    times(&start_planner_time);

    //Effettiva chiamata
    ret_value = system (commline);

    //La funzione system esce con 2 se l'utente killa il planner
    if (ret_value == 2)
    {
	printf ("\nThe external planner was probably killed. Exiting...\n\n");
	exit(2);
    }

    //Tempo finale
    times(&end_planner_time);
    (*elapsed_time) = ( float ) ( ( end_planner_time.tms_cutime - start_planner_time.tms_cutime +
		  	            end_planner_time.tms_cstime - start_planner_time.tms_cstime ) / 100.0);

    //Somma la differenza al totale
    printf ("Elapsed time: %.2f\n", *elapsed_time);

    return 1;
}


int write_problem_header (char* OrigProbFileName, FILE* ProbFile)
{
    FILE *OrigProblem;
    char Line[4000];
    char *ReturnCode;
    char *endstr,*auxstr;

//Apre il file in lettura e controlla l'esito dell'operazione
    OrigProblem = fopen (OrigProbFileName, "r");
    if (OrigProblem == NULL)
    {
	printf ("\nCan't open file '%s' for reading\n", OrigProbFileName);
	exit (2);
    }

    do
    {
	ReturnCode=fgets(Line, 4000, OrigProblem);
	if ( ReturnCode == NULL)
	{
	    printf ("\nUnexpected end of file '%s' while reading problem header.\n\n", OrigProbFileName);
	    exit(2);
	}
	if ( strstr (Line, ":init") )
	{
//Stampa la parte di questa linea fino a '(:init' escluso.
//Si deve tener conto che ci potrebbero essere spazi tra '(' e ':init'
            endstr =  strstr (Line, ":init");
	    while (endstr[0]!='(')
		if (endstr>Line) endstr --;

	    if ( (endstr==Line) && (endstr[0]!='('))
	    {
		printf("\nDidn't find '(' on line containing ':init', please check file %s.\n\n",OrigProbFileName);
		exit(2);
	    }
	    auxstr=Line;
	     
	    while (auxstr!=endstr)
	    {
		fputc(auxstr[0],ProbFile);
		auxstr++;
	    }
	    fputc('\n',ProbFile);
	    break;
	}
	else
	    fprintf (ProbFile,Line);
	    
    }
    
    while (TRUE);
    return 1;
}

//-------------------------------------------------------------------------------------------------
//write_problem_objects
//descrizione degli oggetti specifici del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//objects_list	Puntatore alla lista degli oggetti
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_objects (FILE* ProbFile, fact_list objects_list)
{
    fact_list i_fct;
    token_list i_tok, first_obj;
    char *typename;
//intestazione dichiarazione oggetti
    fprintf (ProbFile,"\t(:objects\n");

//scorre i tipi degli oggetti: per ogni item della fact_list contiene in testa il nome del tipo e quindi gli oggetti di quel tipo
    for (i_fct = objects_list; i_fct; i_fct = i_fct->next)
    {
	fprintf(ProbFile,"\t\t");
//tiene via la prima stringa (il tipo), che in pddl va messa per ultima
	typename = i_fct->item->item;

//puntatore al primo oggetto di quel tipo
	first_obj = i_fct->item->next;
//scorre e stampa gli oggetti di quel tipo
	for (i_tok = first_obj; i_tok; i_tok = i_tok->next)
	    fprintf (ProbFile,"%s ",i_tok->item);
//stampa il tipo degli oggetti
	fprintf(ProbFile,"- %s\n",typename);
    }
//chiude parentesi dichiarazione oggetti
    fprintf (ProbFile,"\t)\n");
    return 1;
}


//-------------------------------------------------------------------------------------------------
//write_problem_inertias
//descrizione delle relazioni non modificabili da alcun operatore
//PARAMETRI
//ProbFile	Puntatore al file del problema
//inertia_list	Puntatore alla lista degli inertia
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_inertias (FILE* ProbFile, token_list inertia_list)
{
    token_list i_tok;
    char begstr[90];
    int i;

//stampa la lista dei fatti inerziali
    for (i_tok=inertia_list; i_tok; i_tok=i_tok->next)
    {
	strcpy (begstr, i_tok->item);
//Sostituisce '_' con uno spazio
	for (i = 0; begstr[i] != '\0'; i++)
	    if (begstr[i] == '_') begstr[i] = ' ';
	fprintf (ProbFile,"\t\t(%s)\n", begstr);
    }

    return 1;
}


//-------------------------------------------------------------------------------------------------
//write_problem_initial_facts
//descrizione dei fatti iniziali del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//list		lista dei fatti iniziali
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_initial_facts (FILE* ProbFile, token_list list)
{
    token_list i_token;
    char begstr[100];
    int i;

//Per ogni oggetto della lista iniziale, aggiungi una relazione
    for ( i_token = list; i_token; i_token = i_token->next )
    {
	fprintf (ProbFile,"\t\t(");

//Fa una copia dell'oggetto della lista
	strcpy(begstr,i_token->item);

//Sostituisce '_' con uno spazio
	for (i = 0; begstr[i] != '\0'; i++)
	    if (begstr[i] == '_') begstr[i] = ' ';

//Stampa su file la stringa 'pulita'
	fprintf(ProbFile, begstr);

	fprintf (ProbFile,")\n");
    }

//Chiusura parentesi fatti iniziali
    fprintf (ProbFile,"\t)\n");

    return 1;
}


//-------------------------------------------------------------------------------------------------
//write_problem_goal_facts
//descrizione degli obiettivi del problema
//PARAMETRI
//ProbFile	Puntatore al file del problema
//list		Lista dei goal da raggiungere
//RETURN
//sempre 1
//-------------------------------------------------------------------------------------------------
int write_problem_goal_facts (FILE* ProbFile, token_list list)
{
    token_list i_token;
    char begstr[100];
    int i;

//Apertura contesto goals
    fprintf (ProbFile,"\t(:goal\t(and\n");

//Per ogni oggetto della lista...
    for ( i_token = list; i_token; i_token = i_token->next )
    {
//Apre la parentesi per l'oggetto corrente
	fprintf (ProbFile,"\t\t\t(");

//Fa una copia dell'oggetto della lista
	strcpy(begstr,i_token->item);

//Sostituisce '_' con uno spazio
	for (i = 0; begstr[i] != '\0'; i++)
	    if (begstr[i] == '_') begstr[i] = ' ';

//Stampa su file la stringa 'pulita'
	fprintf(ProbFile, begstr);

//Chiude la parentesi
	fprintf (ProbFile,")\n");
    }
//Fine file del problema
    fprintf (ProbFile,"\t\t)\n\t)\n)\n");

    return 1;
}


//-------------------------------------------------------------------------------------------------
//get_plan_size
//legge da file la dimensione del piano
//PARAMETRI
//ResFileName	Nome del file risultato
//*max_time	Dimensione del piano
//RETURN
//1 se il piano  stato trovato, 0 se il file  vuoto (il che indica: piano non trovato)
//-------------------------------------------------------------------------------------------------
int get_plan_size (char* ResFileName, int *max_time)
{
    FILE *ResultFile;
    char auxstring[MAX_LENGTH];

    int curr_lev=-1;
    int time=-1;

    ResultFile = fopen (ResFileName, "r");

    if (ResultFile == NULL)
    {
	printf ("\nget_plan_size: Can't open file %s for reading\n", ResFileName);
	printf ("\n");
	exit (2);
    }

//Se tutto ok, ricerca nell'output la parte in cui viene stampato il piano
    while (fgets (auxstring, 90, ResultFile) != NULL)
    {
	sscanf(auxstring,"%d", &curr_lev);
	if (curr_lev > time) time = curr_lev;
    }

    *max_time = time + 1;

    fclose (ResultFile);

    return (time!=-1);
}

//-------------------------------------------------------------------------------------------------
//insert_plan_from_file
//inserisce il sottopiano leggendolo da file
//PARAMETRI
//start_level	punto di inserimento
//RETURN
//numero di livelli inseriti
//-------------------------------------------------------------------------------------------------
int insert_plan_from_file(int start_level)
{
  int level,i;
  char cStr[MAX_LENGTH],
       ResFileName[MAX_LENGTH];

  FILE* fp;

  sprintf (ResFileName, "%s%s%d.pddl", SIMUL_DIR,RESULT_FILE_PREFIX, count);
  //file di output del planner esterno

  fp = fopen (ResFileName, "r");

  if (fp == NULL)
  {
     printf ("\ninsert_plan_from_file: Can't open file %s for reading\n\n", ResFileName);
     exit (2);
  }

  while (!feof(fp))
  {

     skip_comment(fp);
     fscanf(fp,"%s",cStr);

     if (test_level(cStr))
     {
       level = estract_level(cStr);
       fscanf(fp,"%s",cStr);
     }

     for (i = 0; cStr[i] != '\0'; i++)
     {
	 cStr[i] = toupper (cStr[i]);
	 if (cStr[i] == '\n') cStr[i] = '\0';
     }

     insert_simul_vect(cStr, level+start_level);
  }

  fclose (fp);
  return(level+1);
}


//-------------------------------------------------------------------------------------------------
//get_inertial_fact_list
//ricava la lista dei fatti inerziali
//PARAMETRI
//ini_facts	fatti iniziali, non ancora depurati dei fatti inerziali
//inertia_pred	lista dei predicati inerziali
//RETURN
//la lista dei fatti inerziali
//-------------------------------------------------------------------------------------------------
token_list get_inertial_fact_list (token_list ini_facts, token_list inertia_pred)
{

    token_list result=NULL, temp;
    token_list i_tok, i_pred;

//Scorre la lista dei fatti iniziali alla ricerca di predicati inerziali
    for (i_tok=ini_facts; i_tok; i_tok=i_tok->next)
	for (i_pred = inertia_pred; i_pred; i_pred = i_pred->next)
	    if (strstr (i_tok->item, i_pred->item))
	    {
//Trovato un fatto che prevede un predicato inerziale, lo aggiunge alla lista di output
		temp = ( token_list ) calloc( 1, sizeof( token_list_elt ));
		temp->item = i_tok->item;
		temp->next = result;
		result = temp;
	    }

    return result;
}



// The following functions are for simul vector managing

BOOLEAN find_simul_vect(token name, int level)
{
   simul_action *temp_act;

   for (temp_act=simul_vect[level]; temp_act; temp_act=temp_act->next)
   {
      if (strcmp(temp_act->name, name) == SAME)
     	 return (TRUE);
   }

   return (FALSE);
}


// Insert an action, if not present

void insert_simul_vect(token name, int level)
{
   simul_action *temp_act ;

   if ( !find_simul_vect(name, level) )
   {
      temp_act = ( simul_action* ) calloc( 1, sizeof( simul_action ) );
      temp_act->name = new_token( strlen( name ) + 1 );

      strcpy(temp_act->name, name);
      temp_act->next= simul_vect[level];

      temp_act->op_pointer = NULL;

      temp_act->applyed = FALSE;
      temp_act->result = FALSE;

      simul_vect[level]=temp_act;
   }
}


void remove_simul_vect(simul_action* curr_act)
{
   simul_action* prec_act;

   if ( curr_act == simul_vect[0] )
   {
       simul_vect[0] = curr_act->next;
       free(curr_act);
       prec_act=NULL;
       curr_act=simul_vect[0];

       if (simul_vect[0]==NULL)
         down_simul_vect(0);

   }

   // element in a generic position
   else
   {
       prec_act->next = curr_act->next;
       free(curr_act);
       curr_act = prec_act->next;
   }

}


// Free an entire level of simul vector

void free_simul_vect(int level)
{
   simul_action *temp_act, *temp;

   for(temp_act=simul_vect[level]; temp_act; )
   {
      temp = temp_act;
      temp_act = temp->next;
      free(temp);
   }

   simul_vect[level]=NULL;
}


// All levels from "level" are moved one step up, when we want to insert a new (longer) plan

void up_simul_vect(int level)
{
   int i;

   for (i=simul.plan_length-1; i>=level; i--)
      simul_vect[i+1]=simul_vect[i];

   simul_vect[level]=NULL;
   simul.plan_length++;

}


// All levels from "level" are moved one step down, when we want to insert a new (shorter) plan

void down_simul_vect(int level)
{
   int i;

   if (simul_vect[level])
      free_simul_vect(level);

   for (i=level; i<simul.plan_length-1; i++)
      simul_vect[i]=simul_vect[i+1];

   simul_vect[i]=NULL;

   if (simul.plan_length>1)
      simul.plan_length--;

}


void insert_sub_plan(int init_time, int size)
{
    int i;

    for(i=0; i<size; i++)
    {
	if (simul_vect[init_time+i])
	    free_simul_vect(init_time+i);
    }

    insert_plan_from_file (init_time);
}



void insert_executed_vect(token name, BOOLEAN g_changed)
{
   simul_action *temp_act ;

   temp_act = ( simul_action* ) calloc( 1, sizeof( simul_action ) );
   temp_act->name = new_token( strlen( name ) + 1 );

   strcpy(temp_act->name, name);
   temp_act->next = executed_actions[simul.executed_plan_length];

   temp_act->op_pointer = NULL;

   temp_act->applyed = TRUE;
   temp_act->result = FALSE;
   temp_act->goal_changed = g_changed;

   executed_actions[simul.executed_plan_length] = temp_act;

   printf("\n - executed action (%d) :  '%s' ", simul.executed_plan_length, temp_act->name);

   // copio nella struttura del simulatore
   if (g_changed)
   {
      if (strlen(simul.message)<3000)
      {
         strcat(simul.message,"-G:");
         strcat(simul.message,temp_act->name);
      }
      printf("(GOAL CHANGED !!!)\n");
   }

   else
   {
      if (strlen(simul.message)<3000)
      {
         strcat(simul.message,"-IF:");
         strcat(simul.message,temp_act->name);
      }
      printf("\n");
   }
   simul.executed_plan_length++;
}


// these are for reading/writing plan files

FILE* open_plan_file(char* file_name)
{
  FILE* fp;

  fp = fopen(file_name,"r");

  if(fp==NULL)
  {
    printf("\n SIMUL: can't find plan file %s ! \n\n", file_name);
    exit (2);
  }

  return (fp);
}


int test_level(char* cStr)
{
  int i;

  if(cStr[strlen(cStr)-1] == ':')
     i = 1;
  else
     i = 0;

  return(i);
}


int estract_level(char* cStr)
{
  int n,m;

  n = strlen(cStr) - 1;
  cStr[n]='\n';
  m=atoi(cStr);

  return(m);
}



// Search for comment in the input file : they starts whith '#'

void skip_comment(FILE* fp )
{
  char charact;

  while ( (charact = fgetc(fp)) == '#')
  {
     while ( charact != '\n' )
        charact = fgetc(fp);
  }

}


int init_simul_vect(FILE* fp)
{
  int level;
  char cStr[MAX_LENGTH];

  while (!feof(fp))
  {
     skip_comment(fp);
     fscanf(fp,"%s",cStr);

     if (test_level(cStr))
     {
       level = estract_level(cStr);
       fscanf(fp,"%s",cStr);
     }

     insert_simul_vect(cStr, level);
  }

  return(level+1);
}


// Load input plan file

int load_plan(char* plan_file)
{
  int plan_length;
  FILE* pFile;

  pFile = open_plan_file(plan_file);
  plan_length=init_simul_vect(pFile);
  fclose(pFile);

  return(plan_length);
}


// Store result of simulation process on file

void store_simulation(int levels, char* fact_file_name, double planner_time, double planner1_time)
{

  int i;
  char cNameFile[MAX_LENGTH];
  FILE* fp;

  simul_action *temp_act;

  sprintf(cNameFile,"simulation_%s.SIMUL\0", fact_file_name);

  if((fp = fopen( cNameFile, "w")) == NULL)
  {
      printf("\nERROR: cannot open file ! \n\n");
      exit(1);
  }

  fprintf(fp,"# Random seed:\t%d\n", simul.random_seed);
  fprintf(fp,"# Problem:\t%s\n", fact_file_name);

  if (simul.planner.type == FF_GPG)
  {
     fprintf(fp,"# GPG-time:\t%f\n", planner1_time);
     fprintf(fp,"# FF-time:\t%f\n", planner_time);
  }
  else
  {
     fprintf(fp,"# total planning time:\t%f\n", planner_time);
     fprintf(fp,"# planner:\t%d\n", simul.planner.type);
  }

  fprintf(fp,"# levels:\t%d\n", levels);
  fprintf(fp,"# noise rate:\t%d\n", simul.noise_rate);
  fprintf(fp,"# max number of noisy ops:\t%d\n", simul.max_noisy_ops);
  fprintf(fp,"# number of applyed noisy ops:\t%d\n", simul.applyed_noisy_ops);
  fprintf(fp,"# number of irrelevant noisy ops:\t%d\n", simul.irrelevant_noisy_ops);
  fprintf(fp,"# number of not abblyable noisy ops:\t%d\n\n", simul.not_applyable_noisy_ops);

  /* for all non empty operator levels */
  for ( i=0; i<levels; i++ )
  {
     fprintf(fp,"%3d: ", i);

     /* for all operators in plan */
     for (temp_act = executed_actions[i]; temp_act; temp_act = temp_act->next)
     {
        if ( strstr (temp_act->name, "NOISE-") != NULL )
        {
           if (temp_act->goal_changed)
             fprintf(fp,"\t%s\t(NOISE -> GOAL CHANGED)\n", temp_act->name );
           else
             fprintf(fp,"\t%s\t(NOISE)\n", temp_act->name );
        }
        else
           fprintf(fp,"\t%s\n", temp_act->name );
     }

  }

  fclose(fp);
}

