
#ifndef __SIMULATOR_H
#define __SIMULATOR_H


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <values.h>
#include <errno.h>
#include <math.h>
#include <limits.h>
#include <signal.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>

/* defines */

/* technical defines */
#define MAX_LENGTH 256/* maximal string length */
#define BOOLEAN unsigned short int/* you should know about that one */
#define TRUE 1/* this is true */
#define FALSE 0/* and this is not */
#define SAME 0/* for strcmp */

#define TOKEN_LIST_CONNECTOR "_"/*  */

#define CONNECTORCHAR '_'/* no comment */
#define BIGPRIME 8000977/* prime for hashing */
#define NOOP "noop"/* der blanke wahnsinn */

#define CLEAR 2
#define INIT 1/* for initialising static variables, */
#define EXEC 0/* i.e., better readability of fn calls */
#define SEARCH 1/* readability */
#define COMPLETE 0/* of fn call( in search ) */
#define S_HSIZE 4/* size of tables in sub tree branches */
#define S_HASH 3/* for index calculation */

#define NO_QUANT 0
#define ALL_QUANT 1
#define EX_QUANT 2
#define QUANTIFIERS "-AE" /* no, universal, existential */

#define ANY_PRED 0
#define EQ 1
#define EQ_STR "eq"
#define NOT_EQ 2
#define NOT_EQ_STR "not-eq"
#define NOT_STR "not-"
#define NOT_PRED '!'
#define NOT_EQ_PRED "!eq"
#define NO_MEMORY "\nipp:  SORRY, I RAN OUT OF MEMORY!\n"

/* arbitrary preset value defines */
#define HSIZE 50/* size of graph hashtables */
#define MAX_PLAN 1000/* max #time steps for making graph */
#define MAX_MAX_NODES 65536/* upper limit for specified constant max_nodes */
#define NUMINTS 2048 /* MAX_MAX_NODES / 32; length of bitstring in vertex */

/* NOTE: if you need max_nodes( the max number of facts or ops at
 *       one time step ) to be > 1024, #define MAX_MAX_NODES <2 ** n>
 *       and #define NUMINTS <2 ** ( n-5 )>.
 */
#define MAX_GOALS 200/* max number of goals in an array; just technical */

/* Macro for NULL pointer check after mallocs/callocs */
#define CHECK_MEMORY(ptr) if (!(ptr)){\
  fprintf(stderr,"%s", NO_MEMORY);\
  exit(1);}
#define GRAPH( type ) {bytes_graph += sizeof( type );}
#define MEMO( type ) {bytes_memo += sizeof( type );}
#define ARE_MUTEX( v, w ) ( ( ( v->exclusive_vect[w->uid_block] ) & ( w->uid_mask ) ) > 0 )

/* hash values for gam */
#define G_HSIZE 8192
#define G_HASH 8191

#define MEMO( type ) {bytes_memo += sizeof( type );}
#define GRAPHO( type ) {bytes_o += sizeof( type );}
#define GRAPHF( type ) {bytes_f += sizeof( type );}
#define GRAPHE( type ) {bytes_e += sizeof( type ); numbereff++;}
#define GRAPHK( type ) {bytes_k += sizeof( type ); numberkan++;}
#define EXCL( type ) {bytes_excl += sizeof( type );}

#define NO_INCONS -1
#define INIT_ACT   "initial_action"
#define GOAL_ACT   "goal_action"
#define ALL_NAMES   ""
#define ALL_LEVELS  -10
#define TEMP_DELETED -1
#define FIRST_1 0x80000000

#define MAX_NOISE 1000
#define INITIAL 0
#define GOAL    1
#define INIGOAL 2
#define CONNECTION "CONN"
#define GPG     0
#define FF      1
#define IPP4    2
#define FF_GPG  10

#define GPG_COMM_LINE "adj/adj"
#define FF_COMM_LINE "adj/ff/ff"
#define IPP4_COMM_LINE "adj/ipp/ipp"

#define SEARCH 1
#define OPTIMIZE 2

//Possibili limitazioni imposte al planner esterno
#define NO_LIMITS   0
#define CPU_TIME    1
#define LEVELS_NUM  2
#define ACTIONS_NUM 3

#define PROBLEM_FILE_PREFIX "problem"
#define RESULT_FILE_PREFIX "result"
#define PLAN_FILE_PREFIX "plan"


struct tms start_time, search_start;
float build_ad_time, fixpoint_time;


struct PLANNER
{
   int type;  // GPG, FF, IPP4, FF_GPG
   char command_line[MAX_LENGTH];
   char command_line1[MAX_LENGTH];
};


struct SIMUL
{
   struct PLANNER planner;

   int plan_length;
   int executed_plan_length;
   int remaining_plan_length;

   int num_inc;
   int num_actions;
   int random_seed;

   int noise_rate;
   int goal_change_rate;
   int conn_change_rate;
   int noise_change;
   int num_noise_initial_facts;
   int num_noise_goals;
   char message[3000];

   int max_noisy_ops;
   int applyed_noisy_ops;
   int irrelevant_noisy_ops;
   int not_applyable_noisy_ops;
   int goal_heuristics;

   float total_execution_time;

} simul;


/* data structures */

/* the following make up the graph structure for ordering goals:
 * we have a list op_nodes that contains all actions and a
 * hashtable goal_table that contains all the goals that are
 * added by them. the whole stuff is connected by pointers so that
 * every needed information is accessible easily.
 */
typedef struct OPEDGE *o_edge_list, o_edge_list_elt;
typedef struct EFEDGE *e_edge_list, e_edge_list_elt;
typedef struct GOEDGE *g_edge_list, g_edge_list_elt;
typedef struct EFNODE *e_node_list, e_node_list_elt;

typedef struct OPEXCL *o_excl_list, o_excl_list_elt;
typedef struct GOEXCL *g_excl_list, g_excl_list_elt;

typedef struct GONODE *g_node_list, g_node_list_elt;
typedef struct OPNODE *o_node_list, o_node_list_elt;


/* obviously a string */
typedef char *token;

/* GPG string list */
typedef struct TOKENLIST
{

  token item;
  g_node_list go_pointer;

  struct TOKENLIST *next;

} *token_list, token_list_elt;


typedef struct SIMUL_ACTION
{
   token name;

   BOOLEAN applyed;
   BOOLEAN result;
   BOOLEAN goal_changed;

   o_node_list op_pointer;

   struct SIMUL_ACTION* next;

} simul_action;



// struct added for planning with gpg

typedef struct GPG_STRUCT
{

  BOOLEAN is_true;
  BOOLEAN is_precond;
  BOOLEAN is_added;
  BOOLEAN is_deleted;
  short int is_goal;

} gpg_struct;


// GPG main structure for goals and facts

struct GONODE
{

  token name;

  g_edge_list false_facts;
  g_edge_list true_facts;

  int Tcount;
  int Fcount;
  BOOLEAN is_true;
  
  int st;
  int hits;

  int uid_block;
  unsigned int uid_mask;

   /* these are for intermediates */
  BOOLEAN is_initial;
  int count;

  /* and these are the actions and effects that have to do with this goal */
  o_edge_list preconds;
  e_edge_list effconds;
  e_edge_list add_edges;
  e_edge_list del_edges;

  gpg_struct gpg;

  struct GONODE *next;
};


/* an action */
struct OPNODE
{

  token name;/* as usual */
  g_edge_list preconds;/* the preconds; ( pointers to goals in the fact table ) */
  e_node_list effects;/* the effects */

  int uid_block;
  unsigned int uid_mask;

  BOOLEAN is_noop;
  BOOLEAN is_used;

  struct OPNODE *next;
};


/* an effect;
 * in contradiction to the graphplan graph for planning,
 * in GAM we have separated effect nodes;
 * seems to be more natural for ipp style operators
 */
struct EFNODE
{

  BOOLEAN on;/* says if this effect can be used currently; see compute.c */
  o_node_list op;/* the op it belongs to */

  g_edge_list conditions;/* should be self explanatory */
  g_edge_list add_edges;
  g_edge_list del_edges;

  /* these for fixpoint */
  BOOLEAN is_new;
  token_list del_names;

  struct EFNODE *next;
};


/* list of string list */
typedef struct FACTLIST
{

  token_list item;
  struct FACTLIST *next;

} *fact_list, fact_list_elt;


/* effectlist of uninstantiated operator; format:
 * forall <quantified variables> : <conditions> => <add_effects>,<del_effects>
 */
typedef struct EFFECTLIST
{

  fact_list quantified_variables;
  fact_list conditions;
  fact_list add_effects;
  fact_list del_effects;

  struct EFFECTLIST *next;

} *effect_list, effect_list_elt;


/* uninstantiated operator, read from inputfile <name>.ops
 * into this structure
 */
typedef struct OP
{
  token name;

  fact_list params;
  fact_list preconds;
  effect_list effects;

  fact_list params_objects;

  struct OP *next;
} *op_list, op_list_elt;


/* effectlist for instantiated operator; format:
 *   <conditions> => <add_effects>, <del_effects>
 *  where the fact_list s of uninstantiated operators are replaced
 *  by token_lists: connect the elements of a token_list by '_'
 *  to make them be strings and get the resulting token_list as the
 *  list of those strings.
 *  example: at a b on c d changes to at_a_b on_c_d
 *  see token_list_from_fact_list in file util_inst.c
 */
typedef struct INSTEFFECTLIST
{

  token_list conditions;
  token_list add_effects;
  token_list del_effects;

  struct INSTEFFECTLIST *next;

} *inst_effect_list, inst_effect_list_elt;


/* completely instantiated operator, corresponds to actions in the paper */

typedef struct OPERATOR
{
  token name;

  token_list preconditions;
  inst_effect_list effects;

  token_list objects;

  struct OPERATOR *next;
} *operator_list, operator_list_elt;


/* edge lists for everyone... */
struct OPEDGE
{
  o_node_list endpt;
  struct OPEDGE *next;
};

struct EFEDGE
{
  e_node_list endpt;
  struct EFEDGE *next;
};

struct GOEDGE
{
  g_node_list endpt;
  struct GOEDGE *next;
};

struct OPEXCL
{
  char *name;
  int exclusive_vect[NUMINTS];
};

struct GOEXCL
{
  char *name;
  int exclusive_vect[NUMINTS];
};


/* this makes up the hashtable for GAM goals */
typedef g_node_list g_table_type[G_HSIZE];
typedef o_node_list o_table_type[G_HSIZE];

/* just for improving readability of some algorithms */
typedef g_node_list g_array_type[MAX_GOALS];

/* these are technical for intermediates,
 * i.e. to see if there has been a circle: compare bit strings */
typedef int bit_string[NUMINTS];

typedef struct BITSTRINGLIST
{
  bit_string string;

  struct BITSTRINGLIST *prev;
  struct BITSTRINGLIST *next;
} *bit_string_list, bit_string_list_elt;


/* function prototypes */

/* in file utilities.c */

token_list new_token_list(void);
effect_list new_effect_list(void);
token_list copy_token_list(token_list source);
operator_list copy_operator_list( operator_list source );
token new_token(int size);
operator_list new_ground_operator( token name );
void print_operators( operator_list operators );
token_list remove_token(token_list l, token t);
void free_token_list( token_list tl );
token_list token_list_from_string (char* string);


/* in file simulator.c */

int hash_index( token t );
g_node_list insert_go( token_list t );
g_node_list lookup_go( token t );
o_node_list lookup_op( token t );
g_node_list quick_lookup_go( token_list tl );
o_node_list quick_lookup_act( simul_action* act );

void print_g_table(void);
void print_o_table(void);
void initialise_go_op_nodes();
void initialize_gpg_fixpoint(void);
int forward_simulation();
void save_current_state(token_list* tl);
void restore_current_state(token_list tl);
void remove_empty_steps(int max_time);
void plan_check(void);

void print_plan(int time_init, int time_end);
void print_facts_and_goals(int curr_init_time, int curr_max_time);
void print_list(token_list list);

BOOLEAN applicable( token_list list );
g_edge_list insert_g_edge( g_node_list g, g_edge_list list );
o_edge_list insert_o_edge( o_node_list o, o_edge_list list );
e_edge_list insert_e_edge( e_node_list e, e_edge_list list );
void print_ef_node( e_node_list e );
void print_go_op_nodes( void );
void set_gam_go_uid( g_node_list g , int id );
void set_gam_op_uid( o_node_list o , int id );

void initialize_simul_params();
int find_new_plan (int *, fact_list, token_list, token_list, token_list, char* fct_file);
int insert_plan_from_file (int start_level);
int generate_pddl_input (char* DomFileName, char* ProbFileName, fact_list, token_list, token_list facts, token_list goals);
int generate_plan_to_adapt (char* PlanFileName);
int call_planner (char* CommandLine, float* elapsed_time);

int write_problem (FILE* , fact_list , token_list , token_list , token_list goals);
int write_problem_header (char* OrigProbFileName, FILE* ProbFile);
int write_problem_objects (FILE*, fact_list);
int write_problem_inertias (FILE*, token_list);
int write_problem_initial_facts (FILE*, token_list);
int write_problem_goal_facts (FILE*, token_list);

int get_plan_size (char* , int*);

token_list get_inertial_fact_list (token_list, token_list );

int actions_of_level (int );
int count_subplan_actions (int* , int ,int  );

BOOLEAN find_simul_vect(char* name, int level);
void insert_simul_vect(char* name, int level);
void free_simul_vect(int level);
void up_simul_vect(int level);
void down_simul_vect(int level);
void update_simul_vect(int curr_init_time, int curr_max_time, int max_time);

void insert_executed_vect(token name, BOOLEAN g_changed);
void remove_simul_vect(simul_action* temp_act);

int load_plan(char *file_plan);
void store_simulation(int levels, char* fact_file_name, double planner_time, double planner1_time);
int test_level(char* cStr);
int estract_level(char* cStr);
void skip_comment(FILE* fp );
void init_g_table();


/* in file change_plan.c */
void change_plan(int num_noise_initial_facts, int num_noise_goals, fact_list objects_list,
                 token_list inertia_list, token_list f_list, token_list g_list, char* fct_file);

/* in file dijkstra.c */
BOOLEAN dijkstra(char* da, char* a);


/* global variables */


/* technical variables */

extern char *act_filename;
extern int lineno;
extern inst_effect_list help_inst_effect_list;/* used for del_ALL_quantifiing */
extern BOOLEAN minimize;/* has to do with recursive op choice in fn search */

/* globally important variables */

/* domain constants */
extern op_list loaded_ops;/* loaded, uninstantiated operators */
extern operator_list operators;/* list of instantiated operators */
extern token_list initial_facts;/* the initial facts */
extern effect_list raw_goal_facts;/* not yet preprocessed goal facts */
extern token_list goal_facts;/* the goal facts */
extern fact_list orig_constant_list;/* to store all typed objects */
extern fact_list orig_initial_facts;/* stores initials as fact_list */
extern fact_list negated_preds;/* needed for NOT preprocessing */
extern fact_list types;/* fast access to all objects of one type */

/* option preset values */
int max_nodes;/* max number of nodes at one level */
int display_info;/* level of run time info printed */

extern long bytes_graph;
extern long bytes_memo;


/* globally SIMULATOR important variables */

char SIMUL_COMM_LINE[MAX_LENGTH];
char SIMUL_COMM_LINE1[MAX_LENGTH];
char TEMP_DIR[MAX_LENGTH];
char SIMUL_DIR[MAX_LENGTH];

token_list inertial_facts;

simul_action* simul_vect[MAX_PLAN+1];
simul_action* executed_actions[MAX_PLAN+1];

token_list actual_initial_facts, actual_goal_facts;

operator_list noisy_operators;


/* option preset values */

int debug;/* level of debugging info */

/* compute stuff */
o_table_type o_table;
g_table_type g_table;

int gam_total_num_ops;
int gam_total_num_goals;

BOOLEAN gam_fixpoint_reached;
int gam_fixpoint_level;

long bytes_o;
long bytes_f;
long bytes_e;
long bytes_k;
long bytes_excl;
long numbereff;
long numberkan;
long numberexclkan;
long bytes_graph;
long bytes_memo;


#endif //__SIMULATOR_H
