
#include "simulator.h"

/* These are helper functions that might be used everywhere */


token_list new_token_list(void)
{
  token_list result;

  result = ( token_list ) malloc( sizeof( token_list_elt ) );
  
  if (!result)
  {
     fprintf(stderr, "%s", NO_MEMORY); 
     exit(2);
  }

  result->item = NULL;
  result->go_pointer = NULL;
  result->next = NULL;

  return result;
}


effect_list new_effect_list(void)
{
  effect_list result;

  result = ( effect_list ) malloc( sizeof( effect_list_elt ) );

  if (!result)
  {
     fprintf(stderr, "%s", NO_MEMORY);
     exit(2);
  }

  result->quantified_variables = NULL;
  result->conditions = NULL;
  result->del_effects = NULL;
  result->add_effects = NULL;

  return result;
}


token new_token(int size)
{
  char* tok;

  tok = (char*) calloc(size, sizeof(char));
  
  if (!tok)
  {
    fprintf(stderr, "%s", NO_MEMORY);
    exit(2);
  }
  
  return tok;
}


operator_list new_ground_operator( token name )
{
  operator_list act_op;

  act_op = ( operator_list ) malloc( sizeof( operator_list_elt ) );
  act_op->name = new_token( strlen( name ) + 1 );
  strcpy( act_op->name, name );

  act_op->preconditions = NULL;
  act_op->effects = NULL;
  act_op->objects = NULL;
  act_op->next = NULL;
  return act_op;
}


/*
 * This function copies a token_list by creating a new
 * token_list_elt for every element in source and
 * copying the item and next pointer to the new token_list_elt.
 * Then the function calls itself with the next element of source.
 *
 * returns: the start of the new token_list
 */
token_list copy_token_list( token_list source )
{
  token_list temp;

  if ( !source ) temp = NULL;

  else
  {
     temp = new_token_list();
     temp->item = new_token( strlen( source->item ) + 1 );
      
     strcpy( temp->item, source->item );
     temp->go_pointer = source->go_pointer;
     temp->next = copy_token_list( source->next );
  }
  return temp;
}


operator_list copy_operator_list( operator_list source )
{
  operator_list temp;

  if ( !source ) temp = NULL;

  else
  {
     temp = new_ground_operator( source->name );
     temp->preconditions = source->preconditions;
     temp->effects = source->effects;
     temp->objects = source->objects;
     temp->next = copy_operator_list( source->next );
  }
  return temp;
}


/* prints list of instantiated operators to stdout */

void print_operators( operator_list operators )
{

  operator_list i_operator;
  inst_effect_list i_effect;
  token_list i_token;

  for ( i_operator = operators; i_operator; i_operator = i_operator->next )
  {
    printf( "%s\n", i_operator->name );
    printf( "preconditions:\n" );
    for ( i_token = i_operator->preconditions; i_token;i_token=i_token->next)
      printf( "%s ", i_token->item );
    printf( "\neffects:\n" );
    for ( i_effect = i_operator->effects; i_effect;i_effect=i_effect->next)
    {
      for ( i_token = i_effect->conditions; i_token; i_token = i_token->next )
        printf( "%s ", i_token->item );
      if ( i_effect->add_effects )
      {
        printf( "=> ADD " );
        for ( i_token = i_effect->add_effects; i_token;i_token=i_token->next)
          printf( "%s ", i_token->item );
        if ( i_effect->del_effects )
        {
          printf( "DEL " );
          for(i_token = i_effect->del_effects; i_token;i_token=i_token->next)
            printf( "%s ", i_token->item );
        }
      }
      else
      {
        printf( "=> DEL " );
        for ( i_token = i_effect->del_effects; i_token;i_token=i_token->next)
          printf( "%s ", i_token->item );
      }
      printf( "\n" );
    }
    printf( "\n" );
  }
}


/* simply remove a token from a token list */
token_list remove_token(token_list l, token t)
{

  token_list temp_token, previous;

  for (temp_token=l; temp_token; )
  {

     if ( strcmp(temp_token->item, t) == SAME )
     {

	if (temp_token==l)     // first element
	{
	   l=temp_token->next;	
	}
	else
	{
	   previous->next=temp_token->next;
	}
	
 	free (temp_token->item);
//        free (temp_token);
        return (temp_token->next);
     }

     previous=temp_token;
     temp_token=temp_token->next;
  }
}


void free_token_list( token_list tl )
{
  token_list temp;

  while ( tl )
  {
     temp = tl->next;
     free( tl->item );
//     tl->go_pointer = NULL;
//     free( tl );
     tl = temp;
  }

  tl = NULL;
}


void free_effects_list( inst_effect_list eff )
{
  inst_effect_list temp;

  while ( eff )
  {
     temp = eff->next;
     free_token_list( eff->conditions );
     free_token_list( eff->add_effects );
     free_token_list( eff->del_effects );
     eff = temp;
  }

  eff = NULL;
}


void free_operator_list( operator_list op )
{
  operator_list temp;

  while ( op )
  {
     temp = op->next;
     free( op->name );
     free_token_list( op->preconditions );
     free_effects_list ( op->effects );
     free_token_list ( op->objects );
     op = temp;
  }

  operators = NULL;
}


//Genera una nuova stringa, copia dei caratteri da start a end inclusi
char* mids (char* source, int start_c, int end_c)
{
  int i;
  char *result;
  result = new_token(end_c - start_c +3);
  for (i=0; i < end_c - start_c + 1; i++)
    result [i] = source [start_c+i];
  result[end_c - start_c +2]='\0';
  return result;
}                                                                               


token_list token_list_from_string (char* string)
{
  char* charptr;
  token_list temp=NULL,base=NULL,before=NULL;
  char* end;
  char* temp_string;
  charptr=string;
 
  while (charptr!=NULL)
  {
    end = strchr(charptr,'_');
    temp = new_token_list();
    if (!base)
      base=temp;
    if (end)
    {
      temp_string = mids (charptr,0,(int)(end-charptr-1));
      temp->item = temp_string;
      if (before)
        before->next = temp;
    }
    else
    {
      temp_string = new_token (strlen(charptr)+1);
      strcpy (temp_string,charptr);
      temp->item = temp_string;
      if (before)
        before->next = temp;
      break;
    }
    charptr=++end;
    before=temp;
  }
  return base;
}       

